####################################################
# Semantic analyzer to LIE++ Language
# TEC - FOD
# Authors:
#   Michael Gonzalez Rivera
#   Luis Saborio Hernandez
# Edited:
#   Ellioth Ramirez
####################################################

import sys

from LIEparser.parserErrorStrategy import *
from LIEsemantic.semanticErrors import *
# from LIEsemantic.semanticErrorStrategy import *

# ANTLR4 imports
from antlr4 import *
from fodcatListener import fodcatListener
from fodcatLexer import fodcatLexer
from fodcatParser import fodcatParser

## It is informed when ocurred a semantic error and throw the message to standard error.
class SemanticErrorListener():
    
    ## throw the error message to standard error.
    # @param line Line number where the error was detected
    # @param comlun Character position where the error was detected
    # @param msg Error message
    def semanticError(self,  line, column, msg):
        print("linea " + str(line) + ":" + str(column) + " " + msg, file=sys.stderr)

## Semantic analyzer.
# 
# It uses syntax tree (AST) and functions and variables dictionaries to check whether the 
# given program is semantically consistent with language definition.
class SemanticAnalyzer(fodcatListener):
    
    ## Semantic analyzer constructor. Initialize all needed global variables for the
    # succesfully execution of the semantic analyzer.
    def __init__(self):

        ## global_var_dic Identifiers list of all global varaibles.
        # Each item structure var_id: is_array
        # Each item has bool type to know if is an array variable
        self.global_var_dic = {}
        ## This list is cleared when exit declaration function node.
        self.local_var_dic = {}
        ## Each item structure is compund of a key: Function identifier, and
        # a value: wich is a dictionary to. compound by a key: the quantity of params
        # a value: an array, with bool values indicating if the parameter is array or not.
        self.func_dic = {}
        self.func_dic["leer"] = { 0: [] }
        ## Flag indicating if is need to validate if the entered identifier exists.
        self.validate_identifier = False
        ## Flag indicating if we are in a globla or a local scope.
        self.is_global = True
        ## Pointer to semantic error handler. SemanticErrorStrategy is 
        # the object assing by default.
        self.errorHandler = SemanticErrorStrategy()
        ## Pointer to semantic error listener. @link lang_tools.parser.parserErrorStrategy.SemanticErrorListener is 
        # the object assing by default.
        self.listeners = [SemanticErrorListener()]

    ## Enter a parse tree produced by fodcatParser#main_function.
    # @param ctx Rule context.
    def enterMain_function(self, ctx:fodcatParser.Main_functionContext):
        self.is_global = False

    ## Exit a parse tree produced by fodcatParser#main_function.
    # @param ctx Rule context.
    def exitMain_function(self, ctx:fodcatParser.Main_functionContext):
        self.is_global = True


    ## Notify all error registered listeners when an error is detected.
    # @param msg String message to be sended.
    # @param e Error object.
    def notifyErrorListeners(self, msg:str, e:SemanticError):
        for listener in self.listeners:
            listener.semanticError(e.line, e.column, msg)

    ## Remove all errror listeners from listeners list.
    def removeErrorListeners():
        listeners.clear()

    ## Add an error listners to listeners list.
    # @param listener Error listener to be added.
    def addErrorListener(listener:SemanticErrorListener):
        listeners.append(listener)

    ## Enter a parse subtree produced by expression rule.
    # Activate validate identifier flag.
    # @param ctx Rule context
    def enterExpression(self, ctx:fodcatParser.ExpressionContext):
        self.validate_identifier = True

    ## Exit a parse subtree produced by expression rule.
    # Desactivate validate identifier.
    # @param ctx Rule context
    def exitExpression(self, ctx:fodcatParser.ExpressionContext):
        self.validate_identifier = False

    ## Exit a parse subtree produced by assignment rule.
    # Adds the assignation left side to the global or local list as appropriate.
    # @param ctx Rule context
    def enterAssignment(self, ctx:fodcatParser.AssignmentContext):
        if (isinstance(ctx.getChild(0), fodcatParser.IdentifierContext)):
            var_name = ctx.getChild(0).getText()
            
            if (self.is_global):
                if (var_name not in self.global_var_dic):
                    self.global_var_dic[var_name] = False
                elif (self.global_var_dic[var_name] == True):
                    self.errorHandler.reportSemanticError(self, ArrayWithoutIndex( \
                        ctx.start.line, ctx.start.column, var_name))
            else:
                if (var_name not in self.local_var_dic):
                    self.local_var_dic[var_name] = False
                elif (self.local_var_dic[var_name] == True):
                    self.errorHandler.reportSemanticError(self, ArrayWithoutIndex( \
                        ctx.start.line, ctx.start.column, var_name))


    ## Enter a parse tree produced by fodcatParser#array_declaration.
    # Validate if the identifier to array name is avaiable, if it is, the add
    # the array to global_var_dic o local_var_dic as appropiated.
    # If it is not avaible it throw a semantic error
    # @param ctx Rule context
    def enterArray_declaration(self, ctx:fodcatParser.Array_declarationContext):
        var_name = ctx.getChild(0).getText()

        if (self.is_global):
            if (var_name not in self.global_var_dic):
                self.global_var_dic[var_name] = True
            else:
                self.errorHandler.reportSemanticError(self, VariableRedefinition( \
                ctx.start.line, ctx.start.column, var_name))
        else:
            if (var_name not in self.local_var_dic):
                self.local_var_dic[var_name] = True
            else:
                self.errorHandler.reportSemanticError(self, VariableRedefinition( \
                ctx.start.line, ctx.start.column, var_name))


    ## Enter a parse subtree produced by identifier rule.
    # Notify of an error if there are a reference to an identifier wich does not exist.
    # @param ctx Rule context
    def enterIdentifier(self, ctx:fodcatParser.IdentifierContext):
        if (self.validate_identifier \
            and ctx.getText() not in self.local_var_dic \
            and ctx.getText() not in self.global_var_dic \
            and not isinstance(ctx.parentCtx, fodcatParser.Func_callContext)):
            
            self.errorHandler.reportSemanticError(self, NoMatchIdentifier( \
                ctx.start.line, ctx.start.column, ctx.getText()))

    
    ## Enter a parse tree produced by fodcatParser#array_call.
    # Throw a semantic error, if the array called was not defined before
    # @param ctx Rule context
    def enterArray_call(self, ctx:fodcatParser.Array_callContext):
        var_name = ctx.getChild(0).getText()
        if (self.is_global):
            if (var_name not in self.global_var_dic):
                self.errorHandler.reportSemanticError(self, ArrayNotDefined( \
                    ctx.start.line, ctx.start.column, var_name))
            elif (self.global_var_dic[var_name] == False):
                self.errorHandler.reportSemanticError(self, ArrayNotDefined( \
                    ctx.start.line, ctx.start.column, var_name))
        else:
            if (var_name in self.local_var_dic):
                if (self.local_var_dic[var_name] == False):
                    self.errorHandler.reportSemanticError(self, ArrayNotDefined( \
                        ctx.start.line, ctx.start.column, var_name))
            elif (var_name in self.global_var_dic):
                if (self.global_var_dic[var_name] == False):
                    self.errorHandler.reportSemanticError(self, ArrayNotDefined( \
                        ctx.start.line, ctx.start.column, var_name))
            else:
                self.errorHandler.reportSemanticError(self, ArrayNotDefined( \
                    ctx.start.line, ctx.start.column, var_name))
            
    ## Enter a parse subtree produced by function_declaration rule.
    # Advice we are entering to an local scope.
    #
    # Notify of an error if the function is already added. In other words the function the
    # function definition is repeated. 
    # The function firm is compound of the function name and the quantity of parameters.
    # @param ctx Rule context
    def enterFunction_declaration(self, ctx:fodcatParser.Function_declarationContext):
        self.is_global = False

        params = ctx.getChild(3)
        params_type = []
        params_qty = 0
        i = 0
        while (i < params.getChildCount()):
            if (isinstance(params.getChild(i), fodcatParser.IdentifierContext)):
                params_qty += 1
                if(i+1 < params.getChildCount()):
                    if(params.getChild(i+1).getText() != '['):
                        self.local_var_dic[params.getChild(i).getText()] = False
                        params_type.append(False)
                    else:
                        self.local_var_dic[params.getChild(i).getText()] = True
                        params_type.append(True)
                else:
                    self.local_var_dic[params.getChild(i).getText()] = False
                    params_type.append(False)
            i += 1


        func_name = ctx.getChild(1).getText()
        if (func_name in self.func_dic):
            if (params_qty in self.func_dic[func_name]):
                self.errorHandler.reportSemanticError(self, FunctionRedefinition( \
                    ctx.start.line, ctx.start.column, func_name, params_qty))
            else:
                self.func_dic[func_name][params_qty] = params_type
        else:
            self.func_dic[func_name] = { params_qty: params_type }

                
    ## Exit a parse subtree produced by function_declaration rule.
    # Advice we are entering to a global scope. Reason why clear local variables list.
    #
    # The firm is compound of the function name and the quantity of parameters.
    # @param ctx Rule context
    def exitFunction_declaration(self, ctx:fodcatParser.Function_declarationContext):
        self.is_global = True
        self.local_var_dic.clear()


    ## Exit a parse subtree produced by function_declaration rule.
    # Throw an error if the function wich is beeing calling does not exist.
    # Throw an error if the parameters type do not match
    # @param ctx Rule context
    def enterFunc_call(self, ctx:fodcatParser.Func_callContext):
        # getting parameter quantity
        params_qty = ctx.getChild(2).getChildCount()
        if (params_qty > 1):
            params_qty = int(((params_qty - 1) / 2) + 1)

        func_name = ctx.getChild(0).getText()

        if (func_name in self.func_dic):
            if (params_qty in self.func_dic[func_name]):
                params = self.func_dic[func_name][params_qty]
                
                i = 0 #params array index
                j = 0 #identifier child index
                params_id = ctx.getChild(2)
                while i < params_qty:
                    if (isinstance(params_id.getChild(j).getChild(0), fodcatParser.IdentifierContext)):
                        identifier = params_id.getChild(j).getText()
                        if (identifier in self.local_var_dic):
                            if (self.local_var_dic[identifier] != params[i]):
                                if (self.local_var_dic[identifier] == True):
                                    expected_type = "escalar"
                                    getted_type = "arreglo"
                                else:
                                    expected_type = "arreglo"
                                    getted_type = "escalar"
                                
                                self.errorHandler.reportSemanticError(self, NoCompatibleParam( \
                                ctx.start.line, ctx.start.column, expected_type, getted_type))

                        elif (identifier in self.global_var_dic):
                            if (self.global_var_dic[identifier] != params[i]):
                                if (self.global_var_dic[identifier] == True):
                                    expected_type = "arreglo"
                                    getted_type = "escalar"
                                else:
                                    expected_type = "escalar"
                                    getted_type = "arreglo"
                                
                                self.errorHandler.reportSemanticError(self, NoCompatibleParam( \
                                ctx.start.line, ctx.start.column, expected_type, getted_type))

                    j += 2
                    i += 1
            else:
                self.errorHandler.reportSemanticError(self, NoMatchedFunction( \
                ctx.start.line, ctx.start.column, func_name, params_qty))
        else:
            self.errorHandler.reportSemanticError(self, NoMatchedFunction( \
            ctx.start.line, ctx.start.column, func_name, params_qty))


    ## Exit a parse subtree produced by gpio_declaration rule.
    # Throw an error if the parameters to declare the gpio variable are not right.
    # @param ctx Rule context
    def enterGpio_declaration(self, ctx:fodcatParser.Gpio_declarationContext):
        gpio_type = ctx.getChild(2).getText()
        params_qty = ctx.getChild(4).getChildCount()
        if (params_qty > 1):
            params_qty = int(((params_qty - 1) / 2) + 1)

        if (gpio_type == 'boton'):                
            if (params_qty != 1):
                self.errorHandler.reportSemanticError(self, NoMatchedFunction( \
                    ctx.start.line, ctx.start.column, gpio_type, params_qty))

        elif (gpio_type == 'led'):
            if (ctx.getChild(3).getText() == '.'):
                params_qty = ctx.getChild(6).getChildCount()       
                if (params_qty > 1):
                    params_qty = int(((params_qty - 1) / 2) + 1)
                gpio_type = ctx.getChild(4).getText()
            
            if (params_qty != 1):
                self.errorHandler.reportSemanticError(self, NoMatchedFunction( \
                    ctx.start.line, ctx.start.column, gpio_type, params_qty))

        elif (gpio_type == 'servo'):
            if (params_qty != 1 and params_qty != 3):
                self.errorHandler.reportSemanticError(self, NoMatchedFunction( \
                    ctx.start.line, ctx.start.column, gpio_type, params_qty))

## Error messeages definition.
class SemanticErrorStrategy():
    errors = []
    # Report an error for an array call to an array wich is no beeing defined.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.

    def reportArrayNotDefined(self, recognizer: SemanticAnalyzer, e: ArrayNotDefined):
        msg = 'El arreglo: "' + e.offended_identifier + '" no ha sido definidio.'
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e)

    ## Report an eror to an array call wich do not specify an index.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportArrayWithoutIndex(self, recognizer: SemanticAnalyzer, e: ArrayWithoutIndex):
        msg = 'El arreglo: "' + e.offended_identifier + \
            '" requiere que se indique una posición.'
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e)

    ## Report an error for a variable redefinition
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportVariableRedefinition(self, recognizer: SemanticAnalyzer, e: VariableRedefinition):
        msg = 'La variable "' + e.duplicated_identifier + \
            '" fue definida anteriormente. \n'
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e)

    ## Report an error for a function redefinition.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportFuctionRedefinition(self, recognizer: SemanticAnalyzer, e: FunctionRedefinition):
        msg = 'Función "' + e.duplicated_identifier + '", con "' + \
            str(e.params_qty) + '" parametros definida anteriormente.'
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e)

    ## Report an error for a function call to a function not defined.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportNoMatchedFunction(self, recognizer: SemanticAnalyzer, e: NoMatchedFunction):
        msg = 'Función "' + e.no_matched_identifier + '" con "' + \
            str(e.params_qty) + '" parametros, no se ha definido.'
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e)

    ## Report an error when sending no compatible param type.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportNoCompatibleParam(self, recognizer:SemanticAnalyzer, e:NoCompatibleParam):
        msg = 'Paso de parametro incompatible. Tipo esperado "' + e.expected_type + '", tipo recibido "'\
            + e.getted_type + '.'
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e)


    ## Report an error for a identifier call wich is not beeing defined.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportNoMatchIdentifier(self, recognizer:SemanticAnalyzer, e:NoMatchIdentifier):
        msg = 'identificador: "' + e.no_matched_identifier + '" no declarado'
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e)

    ## Depending of the semantic error type chose the error handler.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportSemanticError(self, recognizer:SemanticAnalyzer, e:SemanticError):
        if isinstance(e, NoMatchIdentifier):
            self.reportNoMatchIdentifier(recognizer, e)
        elif isinstance(e, FunctionRedefinition):
            self.reportFuctionRedefinition(recognizer, e)
        elif isinstance(e, NoMatchedFunction):
            self.reportNoMatchedFunction(recognizer, e)
        elif isinstance(e, VariableRedefinition):
            self.reportVariableRedefinition(recognizer, e)
        elif isinstance(e, ArrayWithoutIndex):
            self.reportArrayWithoutIndex(recognizer, e)
        elif isinstance(e, ArrayNotDefined):
            self.reportArrayNotDefined(recognizer, e)
        elif isinstance(e, NoCompatibleParam):
            self.reportNoCompatibleParam(recognizer, e)

        sys.exit(0)

def main(argv):
    inp = FileStream(argv[1])
    lexer = fodcatLexer(inp)
    stream = CommonTokenStream(lexer)
    parser = fodcatParser(stream)
    parser._errHandler = ParserErrorStrategy()
    parser.removeErrorListeners()
    parser.addErrorListener(ParserErrorListener())
    tree = parser.program()
    walker = ParseTreeWalker()
    semanticAnalyzer = SemanticAnalyzer()
    walker.walk(semanticAnalyzer, tree)

if __name__ == '__main__':
    main(sys.argv)


####################################################
# Code generator (transpiler) to LIE++ Language
# TEC - FOD
# Authors:
#   Michael Gonzalez Rivera
#   Luis Saborio Hernandez
# Edited:
#   Ellioth Ramirez
####################################################

import sys
import platform
import subprocess
import re
import math

from antlr4 import *
from gen.fodcatListener import fodcatListener
from gen.fodcatLexer import fodcatLexer
from gen.fodcatParser import fodcatParser

from multi.multiFun import multiFun
GTI = multiFun()

# Documentation for CodeGenerator
#
# Compiler code generator that works over the AST generated by ANTLR4.
# It translates the LIE++ code to Python, and also executes, if it can, the translated
# code.


class CodeGenerator(fodcatListener):

    # CodeGenerator constructor. Initialize all needed global variables for the
    # succesfully execution of the code generator.
    # @param output Python output file name.
    def __init__(self, output, canGPIOFlag, input_mode, inputs=[]):
        self.input_mode = input_mode
        self.inputs = inputs
        # Store the Python output file name.
        self.output = output
        # String to store the translated code.
        self.program = ""

        # Identation level counter.
        self.countIndent = 0
        # Used to set a diferent name to each "while" translated form a "repetir"
        # instrucction.
        self.cicleCount = 0

        # Flag to store if it is translating an "if" command.
        self.ifFlag = False
        # Flag to store if it is translating an "else if" command.
        self.elseifFlag = False
        # Flag to store if it is translating a "while" command.
        self.whileFlag = False
        # Flag to store if the function call is part of an expression.
        self.expressionFlag = False
        # Flag to store if the variable/array declaration is global or local.
        self.localsFlag = False  # Set the local definitions flag to false
        # Stores if it can use the GPIOs
        self.canGPIOFlag = canGPIOFlag
        # Stores if buttons have been used in source code
        self.hasButtonFlag = False
        # Stores if LEDs have been used in source code
        self.hasLEDFlag = False
        # Stores if variable brightness LEDs have been used in source code
        self.hasVLEDFlag = False
        # Stores if servos have been used in source code
        self.hasServoFlag = False
        # Stores if angular servos have been used in source code
        self.hasAngularServoFlag = False
        # Stores if sleep instruccion has been used in source code
        self.hasSleepFlag = False
        # Stores if GPIOs have been used in source code
        self.hasGPIOFlag = False
        # Flag to store if the variable/array declarations are using dynamic declarations/assignations.
        self.canExecuteFlag = True

        # Dictionary that stores local variables for the exec() parameters.
        self.locals = {}
        # Dictionary that stores global variables for the exec() parameters.
        self.globals = {}
        # Defines that __builtins__ are not used for the exec() command.
        # self.globals["__builtins__"] = None
        # Stores the definition of the main function.
        self.globals["main"] = 'def main()'
        # ## Specifies that the str function will be used for the exec() command.
        # self.globals["str"] = str
        # ## Specifies that the print function will be used for the exec() command.
        # self.globals["print"] = print

    # Sets the indentation level in program code. It works with the value of the countIndent variable.
    # @param indentLevel Receives the indentation level to be set at program code.
    def indent(self, indentLevel):
        while indentLevel > 0:
            # Add indentation at program code.
            self.program += "\t"
            indentLevel -= 1

    # Transforms keywords and symbols to Python using regex substitution or string replacement.
    # @param String of expression.
    # @return Expression modified to fit Python keywords.
    def processKeyword(self, exp):
        # sub is a subtitution in "alone" spaces, not substrings.
        exp = re.sub(r"\bpi\b", str(math.pi), exp)
        exp = re.sub(r"\beuler\b", str(math.e), exp)
        exp = re.sub(r"\bverdadero\b", "True", exp)
        exp = re.sub(r"\bfalso\b", "False", exp)
        exp = re.sub(r"\by\b", " and ", exp)
        exp = exp.replace("&", " and ")  # replace is in all the string.
        exp = re.sub(r"\s+o\s+", " or ", exp)
        exp = exp.replace("|", " or ")
        exp = re.sub(r"\b(no[(]|no\s+[(])\b", " not (", exp)
        exp = re.sub(r"([!][(]|[!]\s+[(])", " not (", exp)
        exp = re.sub(r"\bno\b", "not", exp)
        exp = re.sub(r"\s+=\s+", " == ", exp)
        exp = re.sub(r"\bdiferente\b", " != ", exp)
        exp = re.sub(r"\bigual\b", " == ", exp)
        exp = re.sub(r"\bmod\b", " % ", exp)

        exp = re.sub(r"\b.encendido\b", ".on", exp)
        exp = re.sub(r"\b.apagado\b", ".off", exp)
        exp = re.sub(r"\b.parpadear\b", ".blink", exp)
        exp = exp.replace(".brillo", ".value")
        exp = re.sub(r"\b.minimo\b", ".min", exp)
        exp = re.sub(r"\b.medio\b", ".mid", exp)
        exp = re.sub(r"\b.maximo\b", ".max", exp)
        exp = exp.replace(".angulo", ".angle")
        return exp

    # Checks if is an expression, searching if there
    # is a logic or arithmetic operator.
    # @param varValue String of code to be checked.
    # @return Boolean items list.
    def checkIfExpression(self, varValue):
        items = list()
        items.append(varValue.__contains__("+"))
        items.append(varValue.__contains__("-"))
        if items[1]:
            if not((varValue.find("-") > 0) | (varValue.count("-") > 1)):  # If is not a negative number
                items[1] = False
        items.append(varValue.__contains__("%"))
        items.append(varValue.__contains__("("))
        items.append(varValue.__contains__("*"))
        items.append(varValue.__contains__("/"))
        items.append(varValue.__contains__(">"))
        items.append(varValue.__contains__("<"))
        items.append(varValue.__contains__(" and "))
        items.append(varValue.__contains__(" or "))
        items.append(varValue.__contains__(" not "))
        items.append(varValue.__contains__("=="))
        return items

    # Checks if is a logic or relational expression, searching if there is a logical or
    # relational operator. Is used for the print translation.
    # @param varValue String of code to be checked.
    # @return Boolean items list.

    def checkIfLogicalRelationalExpression(self, varValue):
        items = list()
        items.append(varValue.__contains__(">"))
        items.append(varValue.__contains__("<"))
        items.append(varValue.__contains__(" and "))
        items.append(varValue.__contains__(" or "))
        items.append(varValue.__contains__(" not "))
        items.append(varValue.__contains__("=="))
        items.append(varValue.__contains__("True"))
        items.append(varValue.__contains__("False"))
        return items

    # ************************************************************************************
    # Initial Rule
    # ************************************************************************************

    # Enter a parse tree produced by fodcatParser#program.
    # @param ctx:fodcatParser.CommandContext
    def enterProgram(self, ctx: fodcatParser.ProgramContext):
        self.program = ""
        self.program += "input_mode = '" + self.input_mode + "'"
        inputs = ''
        if self.inputs != []:
            inputs = '['
            first = True
            for input in self.inputs:
                if first:
                    inputs += input
                    first = False
                else:
                    inputs += ', ' + input
            inputs += ']'
        self.program += "\nentradas = " + inputs

    # Exit a parse tree produced by fodcatParser#program. Writes in a file the Python
    # translated program, and, if there are no problems, it executes program with the
    # specific variables.
    # @param ctx:fodcatParser.CommandContext
    def exitProgram(self, ctx: fodcatParser.ProgramContext):
        self.program += "\ndef leer():\n"
        self.indent(1)
        self.program += "if input_mode == 'automatico':\n"
        self.indent(2)
        self.program += "return entradas.pop(0)\n"
        self.indent(1)
        self.program += "elif input_mode == 'manual':\n"
        self.indent(2)
        self.program += "return input()\n\n"

        self.program += "\nmain()"  # Add the main function call
        self.program += "\nfrom types import FunctionType"
        self.program += "\n\nnombre_variables = " + \
            str(list(self.globals.keys()))
        self.program += "\nvar_globales = {}"
        self.program += "\nfor nombre_variable in nombre_variables:\n"
        self.countIndent += 1  # Add 1 to the indentation level
        self.indent(self.countIndent)
        self.program += "if (type(globals()[nombre_variable]) is not FunctionType and nombre_variable != 'main'):\n"
        self.indent(2)
        self.program += "var_globales[nombre_variable] = globals()[nombre_variable]"
        self.program += "\nprint('dict_globales = ', var_globales)"

        if self.output == None:
            # does nothing
            return

        if self.output != None:
            # Writes in a file the Python translated program
            self.output.write(self.program)
        # If is the correct device to be executed
        if (self.canGPIOFlag and self.hasGPIOFlag) or (not self.hasGPIOFlag):
            # Executes program with the specific variables
            mmFun = multiFun()
            mmFun.printPreatty(mmFun.execCode(self.program))

        # If is using the GPIOs, but is not running over a Raspberry Pi
        elif (not self.canGPIOFlag) and self.hasGPIOFlag:
            print("El código no se ejecutará, ya que se están utilizando los GPIOs,"
                  " pero no se está ejecutando en una Raspberry Pi.\nPuede transferir el archivo \""
                  + self.output.name + "\" a una Raspberry Pi para realizar la ejecución manual.")
        # if self.canExecuteFlag: # If there are no problems
        #     # If is the correct device to be executed
        #     if (self.canGPIOFlag and self.hasGPIOFlag) or (not self.hasGPIOFlag):
        #         #Execute program with the specific variables
        #         exec(self.program, self.globals, self.locals)
        #     # If is using the GPIOs, but is not running over a Raspberry Pi
        #     elif (not self.canGPIOFlag) and self.hasGPIOFlag:
        #         print("El código no se ejecutará, ya que se están utilizando los GPIOs,"
        #               " pero no se está ejecutando en una Raspberry Pi.\nPuede transferir el archivo \""
        #               + self.output.name + "\" a una Raspberry Pi para realizar la ejecución manual.")
        # else: # If there is a dynamic assignation
        #     print("Hubo una asignación dinámica en el código, por lo que no se realizará "
        #           "la ejecución automática, ya que no garantiza un resultado correcto.\n"
        #           "Puede realizar la ejecución manual del archivo \"" + self.output.name + "\".")

    # Enter a parse tree produced by fodcatParser#main_function.
    # @param ctx:fodcatParser.CommandContext
    def enterMain_function(self, ctx: fodcatParser.Main_functionContext):
        self.countIndent += 1  # Add 1 to the indentation level
        self.program += "\ndef main():\n"  # Add the main function

        self.globals["main"] = "def main()"
        self.localsFlag = True  # Set the local definitions flag to true
        if ctx.getChild(4).getText() == "fin":  # If the main function is empty
            self.program += "\tpass"  # Do nothing

    # Exit a parse tree produced by fodcatParser#main_function.
    # @param ctx:fodcatParser.CommandContext
    def exitMain_function(self, ctx: fodcatParser.Main_functionContext):
        self.indent(self.countIndent)
        self.program += "print('dict_locales = ', locals())\n"
        self.countIndent -= 1  # Substract 1 from the indentation level
        self.program += "\n"
        self.localsFlag = False  # Set the local definitions flag to false

    # ************************************************************************************
    # Commands
    # ************************************************************************************

    # Enter a parse tree produced by fodcatParser#command.
    # @param ctx:fodcatParser.CommandContext
    def enterCommand(self, ctx: fodcatParser.CommandContext):
        # If the command is a "mientras" statement
        if ctx.getChild(0).getText() == "mientras":
            self.cicleCount += 1  # Add 1 to the cicleCount variable
            self.indent(self.countIndent)  # Set proper indentation level
            whileExpression = GTI.getTextInterval(ctx, ctx.getChild(2))
            whileExpression = self.processKeyword(
                whileExpression)  # Gets the "while expression"
            if (whileExpression.__contains__(".presionado")):
                # Adds while statement when a button is used
                self.program += "while (" + ctx.getChild(
                    2).getChild(0).getText() + ".is_pressed):\n"
                self.whileFlag = True
            else:
                self.program += "while (" + whileExpression + "):\n"
            self.countIndent += 1  # Add 1 to the indentation level

            if ctx.getChildCount() < 7:  # If the "mientras" body is empty
                self.indent(self.countIndent)  # Set proper indentation level
                self.program += "pass\n"  # Do nothing

        # If the command is a "repetir" statement
        elif ctx.getChild(0).getText() == "repetir":
            self.cicleCount += 1  # Add 1 to the cicleCount variable
            self.indent(self.countIndent)  # Set proper indentation level
            repetitions = GTI.getTextInterval(ctx, ctx.getChild(2))
            # Gets the times the commands inside the "repetir" statement should be repeated
            repetitions = self.processKeyword(repetitions)
            # Creates the counter to be used by the while statement
            self.program += "whileCounter" + \
                str(self.cicleCount) + "=" + repetitions + "\n"
            try:
                self.locals["whileCounter" +
                            str(self.cicleCount)] = int(repetitions)
            except:
                pass
            self.indent(self.countIndent)  # Set proper indentation level
            # Emulates the "repetir" statement with a "while"
            self.program += "while whileCounter" + \
                str(self.cicleCount) + " > 0:\n"
            self.countIndent += 1  # Add 1 to the indentation level
            self.indent(self.countIndent)  # Set proper indentation level
            self.program += "whileCounter" + \
                str(self.cicleCount) + " -= 1\n"  # Decreases by 1 the counter

            if ctx.getChildCount() < 7:  # If the "repetir" body is empty
                self.indent(self.countIndent)  # Set proper indentation level
                self.program += "pass\n"  # Do nothing

        # If the command is a "imprimir" statement
        elif ctx.getChild(0).getText() == "imprimir":
            self.indent(self.countIndent)  # Set proper indentation level
            childCounter = 2
            self.program += "print("
            expression = False  # Flag to check if is an expression
            relLogExpression = False  # Flag to check if is a relational or logical expression
            first = True  # First value to be printed flag
            while (childCounter < ctx.getChildCount()):  # Gets all the elements to be printed
                # Checks if the child is an expression
                if not(ctx.getChild(childCounter).getText().__contains__("\"")):
                    expression = True
                    # Checks if the child is a relational or logical expression
                    if (any(self.checkIfLogicalRelationalExpression(self.processKeyword(ctx.getChild(childCounter).getText())))):
                        relLogExpression = True
                if first:  # If is the first value to be printed
                    first = False
                    if expression:  # If is an expression
                        if relLogExpression:  # If is a relational or logical expression
                            # The "if" statement inside the print statement is to translate to spanish the boolean values
                            self.program += "str(\"verdadero\" if (" + self.processKeyword(
                                ctx.getChild(childCounter).getText()) + ") else \"falso\")"
                        else:
                            self.program += "str(" + self.processKeyword(
                                ctx.getChild(childCounter).getText()) + ")"
                    else:  # If is not an expression
                        self.program += ctx.getChild(childCounter).getText()
                else:  # If is not the first value to be printed
                    if expression:  # If is an expression
                        if relLogExpression:  # If is a relational or logical expression
                            # The "if" statement inside the print statement is to translate to spanish the boolean values
                            self.program += " + str(\"verdadero\" if (" + self.processKeyword(
                                ctx.getChild(childCounter).getText()) + ") else \"falso\")"
                        else:
                            self.program += " + str(" + self.processKeyword(
                                ctx.getChild(childCounter).getText()) + ")"
                    else:  # If is not an expression
                        self.program += " + " + \
                            ctx.getChild(childCounter).getText()
                expression = False  # Clear expression flag
                relLogExpression = False  # Clear relational or logical expression flag
                # Add 2 to the childCounter (to skip the print operator)
                childCounter += 2
            self.program += ")"

        # If the command is a "devolver" statement
        elif ctx.getChild(0).getText() == "devolver":
            self.indent(self.countIndent)  # Set proper indentation level
            self.program += "return "
            if ctx.getChildCount() > 1:  # If is not an empty return
                # Add return value
                self.program += self.processKeyword(
                    GTI.getTextInterval(ctx, ctx.getChild(1)))
        else:
            pass

    # Exit a parse tree produced by fodcatParser#command.
    # @param ctx:fodcatParser.CommandContext
    def exitCommand(self, ctx: fodcatParser.CommandContext):
        # If the command was a "mientras" or "repetir" statement
        if (ctx.getChild(0).getText() == "mientras") | (ctx.getChild(0).getText() == "repetir"):
            self.countIndent -= 1  # Substract 1 from the indentation level
            self.whileFlag = False
        # If the command was a "imprimir" or "devolver" statement
        elif (ctx.getChild(0).getText() == "imprimir") | (ctx.getChild(0).getText() == "devolver"):
            self.program += "\n"
        else:
            pass

    # Enter a parse tree produced by fodcatParser#ifstat.
    # @param ctx:fodcatParser.CommandContext
    def enterIfstat(self, ctx: fodcatParser.IfstatContext):
        if not self.elseifFlag:  # Checks if it is not translating an "else if" command
            self.indent(self.countIndent)  # Set proper indentation level
        else:
            self.elseifFlag = False
        # Get text with whitespace
        ifCommand = GTI.getTextInterval(ctx, ctx)
        # Get text between "si" and "entonces"
        logicExp = str(re.findall("(?<=si)(.*?)(?=entonces)", ifCommand)[0])
        if (logicExp.__contains__(".presionado")):
            # Adds if statement when a button is used
            self.program += "if (" + \
                ctx.getChild(2).getChild(0).getText() + ".is_pressed):\n"
            self.ifFlag = True
        else:
            # Adds if statement to the program after converting keywords
            self.program += "if (" + self.processKeyword(logicExp) + "):\n"
        if ctx.getChildCount() < 6:  # If is an empty "if"
            self.indent(self.countIndent)  # Set proper indentation level
            self.program += "\tpass"  # Do nothing
        self.countIndent += 1  # Add 1 to the indentation level

    # Exit a parse tree produced by fodcatParser#ifstat.
    # @param ctx:fodcatParser.CommandContext
    def exitIfstat(self, ctx: fodcatParser.IfstatContext):
        if (str(re.findall("(?<=si)(.*?)(?=entonces)",
                           GTI.getTextInterval(ctx, ctx))[0]).__contains__(".presionado")):
            self.ifFlag = False
        self.program += "\n"
        self.countIndent -= 1  # Substract 1 from the indentation level

    # Enter a parse tree produced by fodcatParser#elsestat.
    # @param ctx:fodcatParser.CommandContext
    def enterElsestat(self, ctx: fodcatParser.ElsestatContext):
        self.indent(self.countIndent)  # Set proper indentation level
        if ctx.getChildCount() >= 2:  # If the "else" statement has more than 2 children nodes
            if ctx.getChild(1).getChild(0).getText() == "si":  # Is an "else if" statement
                self.elseifFlag = True  # Set elsif flag to true
                self.program += "el"
            else:  # Is an "else" statement
                self.program += "else:\n"
                self.countIndent += 1  # Add 1 to the indentation level

        else:  # If is a regular "else" statement
            self.program += "else:\n"
            self.indent(self.countIndent)  # Set proper indentation level
            self.program += "\tpass\n"

    # Exit a parse tree produced by fodcatParser#elsestat.
    # @param ctx:fodcatParser.CommandContext
    def exitElsestat(self, ctx: fodcatParser.ElsestatContext):
        if ctx.getChildCount() == 2:
            if not(ctx.getChild(1).getChild(0).getText() == "si"):
                self.countIndent -= 1  # Substract 1 from the indentation level
        else:
            self.countIndent -= 1  # Substract 1 from the indentation level

    # Enter a parse tree produced by fodcatParser#func_call.
    # @param ctx:fodcatParser.CommandContext
    def enterFunc_call(self, ctx: fodcatParser.Func_callContext):
        if not self.expressionFlag:  # If the function call is not inside an expression
            self.indent(self.countIndent)  # Set proper indentation level
            self.program += ctx.getText()

    # Exit a parse tree produced by fodcatParser#func_call.
    # @param ctx:fodcatParser.CommandContext
    def exitFunc_call(self, ctx: fodcatParser.Func_callContext):
        if not self.expressionFlag:  # If the function call is not inside an expression
            self.program += "\n"

    # Enter a parse tree produced by fodcatParser#flow_exp.
    # @param ctx:fodcatParser.CommandContext
    def enterFlow_exp(self, ctx: fodcatParser.Flow_expContext):
        self.indent(self.countIndent)  # Set proper indentation level
        # If the command is a "finciclo" statement
        if ctx.getChild(0).getText() == "finciclo":
            self.program += "break"
        # If the command is a "continuar" statement
        elif ctx.getChild(0).getText() == "continuar":
            self.program += "continue"

    # Exit a parse tree produced by fodcatParser#flow_exp.
    # @param ctx:fodcatParser.CommandContext
    def exitFlow_exp(self, ctx: fodcatParser.Flow_expContext):
        self.program += "\n"

    # Enter a parse tree produced by fodcatParser#assignment.
    # @param ctx:fodcatParser.CommandContext
    def enterAssignment(self, ctx: fodcatParser.AssignmentContext):
        self.indent(self.countIndent)  # Set proper indentation level
        if ctx.getChild(0).getChildCount() > 1:  # Its an array assignation
            position = GTI.getTextInterval(ctx, ctx.getChild(0).getChild(2))
            position = self.processKeyword(position)  # Get translated position
            if position in self.locals:
                position = str(self.locals[position])
            elif position in self.locals:
                position = str(self.globals[position])
            expressionListPosition = self.checkIfExpression(
                position)  # Checks if the position is an expression
            value = GTI.getTextInterval(ctx, ctx.getChild(2))
            value = self.processKeyword(value)  # Get translated value
            expressionListValue = self.checkIfExpression(
                value)  # Checks if the value is an expression
            # If position or value is an expression
            if (any(expressionListPosition)) | (any(expressionListValue)):
                if not self.localsFlag:  # If is a global array assignation
                    self.canExecuteFlag = False  # Will not be executed
            else:
                if self.localsFlag & self.canExecuteFlag:  # If is a local array assignation
                    # If the array is local
                    if ctx.getChild(0).getChild(0).getText() in self.locals:
                        # If a value has not been assigned
                        if self.locals[ctx.getChild(0).getChild(0).getText()][int(position)] == None:
                            try:
                                self.locals[ctx.getChild(0).getChild(0).getText()][int(position)] = int(
                                    value)  # Checks if the value is an integer number
                            except:
                                try:
                                    self.locals[ctx.getChild(0).getChild(0).getText()][int(position)] = float(
                                        value)  # Checks if the value is a float number
                                except:
                                    try:
                                        # Checks if the value is a boolean
                                        if value == "True":  # Checks if the value is true
                                            self.locals[ctx.getChild(0).getChild(
                                                0).getText()][int(position)] = True
                                        elif value == "False":  # Checks if the value is false
                                            self.locals[ctx.getChild(0).getChild(
                                                0).getText()][int(position)] = False
                                        else:
                                            if "\"" in value:  # Checks if the value is a string
                                                self.locals[ctx.getChild(0).getChild(
                                                    0).getText()][int(position)] = str(value)
                                            else:  # Is an identifier
                                                self.canExecuteFlag = False
                                    except:
                                        pass
                    else:  # If the array is global
                        # If a value has not been assigned
                        if self.globals[ctx.getChild(0).getChild(0).getText()][int(position)] == None:
                            try:
                                self.globals[ctx.getChild(0).getChild(0).getText()][int(position)] = int(
                                    value)  # Checks if the value is an integer number
                            except:
                                try:
                                    self.globals[ctx.getChild(0).getChild(0).getText()][int(position)] = float(
                                        value)  # Checks if the value is a float number
                                except:
                                    try:
                                        # Checks if the value is a boolean
                                        if value == "True":  # Checks if the value is true
                                            self.globals[ctx.getChild(0).getChild(
                                                0).getText()][int(position)] = True
                                        elif value == "False":  # Checks if the value is false
                                            self.globals[ctx.getChild(0).getChild(
                                                0).getText()][int(position)] = False
                                        else:
                                            if "\"" in value:  # Checks if the value is a string
                                                self.globals[ctx.getChild(0).getChild(
                                                    0).getText()][int(position)] = str(value)
                                            else:  # Is an identifier
                                                self.canExecuteFlag = False
                                    except:
                                        pass
                elif(self.canExecuteFlag):  # If is a global array assignation
                    try:
                        self.globals[ctx.getChild(0).getChild(0).getText()][int(position)] = int(
                            value)  # Checks if the value is an integer number
                    except:
                        try:
                            self.globals[ctx.getChild(0).getChild(0).getText()][int(position)] = float(
                                value)  # Checks if the value is a float number
                        except:
                            try:
                                # Checks if the value is a boolean
                                if value == "True":  # Checks if the value is true
                                    self.globals[ctx.getChild(0).getChild(
                                        0).getText()][int(position)] = True
                                elif value == "False":  # Checks if the value is false
                                    self.globals[ctx.getChild(0).getChild(
                                        0).getText()][int(position)] = False
                                else:
                                    if "\"" in value:  # Checks if the value is a string
                                        self.globals[ctx.getChild(0).getChild(
                                            0).getText()][int(position)] = str(value)
                                    else:  # Is an identifier
                                        self.canExecuteFlag = False
                            except:
                                pass
            self.program += ctx.getChild(0).getChild(0).getText() + \
                " [" + ctx.getChild(0).getChild(2).getText() + "] = " + value
        else:  # Its a variable declaration or value assignation
            value = GTI.getTextInterval(ctx, ctx.getChild(2))
            value = self.processKeyword(value)  # Get translated value
            # Checks if the value is an expression
            expressionList = self.checkIfExpression(value)
            if (any(expressionList)):  # If value is an expression
                if not self.localsFlag:  # If is a global declaration or value assignation
                    self.canExecuteFlag = False  # Will not be executed
            else:
                if self.localsFlag:  # If is a local declaration or value assignation
                    # If is the first definition
                    if (not ctx.getChild(0).getText() in self.locals) & (not ctx.getChild(0).getText() in self.globals):
                        try:
                            # Checks if the value is an integer number
                            self.locals[ctx.getChild(0).getText()] = int(value)
                        except:
                            try:
                                self.locals[ctx.getChild(0).getText()] = float(
                                    value)  # Checks if the value is a float number
                            except:
                                try:
                                    # Checks if the value is a boolean
                                    if value == "True":  # Checks if the value is true
                                        self.locals[ctx.getChild(
                                            0).getText()] = True
                                    elif value == "False":  # Checks if the value is false
                                        self.locals[ctx.getChild(
                                            0).getText()] = False
                                    else:
                                        if "\"" in value:  # Checks if the value is a string
                                            self.locals[ctx.getChild(
                                                0).getText()] = str(value)
                                        else:  # Is an identifier
                                            self.canExecuteFlag = False
                                except:
                                    pass
                else:  # If is a global declaration or value assignation
                    try:
                        self.globals[ctx.getChild(0).getText()] = int(
                            value)  # Checks if the value is an integer number
                    except:
                        try:
                            self.globals[ctx.getChild(0).getText()] = float(
                                value)  # Checks if the value is a float number
                        except:
                            try:
                                # Checks if the value is a boolean
                                if value == "True":  # Checks if the value is true
                                    self.globals[ctx.getChild(
                                        0).getText()] = True
                                elif value == "False":  # Checks if the value is false
                                    self.globals[ctx.getChild(
                                        0).getText()] = False
                                else:
                                    if "\"" in value:  # Checks if the value is a string
                                        self.globals[ctx.getChild(
                                            0).getText()] = str(value)
                                    else:  # Is an identifier
                                        self.canExecuteFlag = False
                            except:
                                pass
            self.program += ctx.getChild(0).getText() + " = " + value

    # Exit a parse tree produced by fodcatParser#assignment.
    # @param ctx:fodcatParser.CommandContext
    def exitAssignment(self, ctx: fodcatParser.AssignmentContext):
        self.program += "\n"

    # Enter a parse tree produced by fodcatParser#gpio.
    # @param ctx:fodcatParser.CommandContext
    def enterGpio(self, ctx: fodcatParser.GpioContext):
        # if not self.ifFlag:
        if not (self.ifFlag | self.whileFlag):
            self.indent(self.countIndent)  # Set proper indentation level
            if ctx.getChild(3).getText() == ":=":
                value = GTI.getTextInterval(ctx, ctx.getChild(4))
                value = self.processKeyword(value)  # Get translated value
                self.program += ctx.getChild(0).getText() + \
                    self.processKeyword(
                        "." + ctx.getChild(2).getText()) + " = " + value
            else:
                self.program += self.processKeyword(ctx.getText())

    # Exit a parse tree produced by fodcatParser#gpio.
    # @param ctx:fodcatParser.CommandContext
    def exitGpio(self, ctx: fodcatParser.GpioContext):
        # if not self.ifFlag:
        if not (self.ifFlag | self.whileFlag):
            self.program += "\n"

    # Enter a parse tree produced by fodcatParser#sleep.
    # @param ctx:fodcatParser.CommandContext
    def enterSleep(self, ctx: fodcatParser.SleepContext):
        self.indent(self.countIndent)  # Set proper indentation level
        if not self.hasSleepFlag:
            self.program = "from time import sleep\n" + \
                self.program  # Import the GPIO class
            self.hasSleepFlag = True
        self.program += "sleep(" + ctx.getChild(2).getText() + ")"

    # Exit a parse tree produced by fodcatParser#sleep.
    # @param ctx:fodcatParser.CommandContext
    def exitSleep(self, ctx: fodcatParser.SleepContext):
        self.program += "\n"

    #
    # ************************************************************************************
    # Expressions
    # ************************************************************************************

    # Enter a parse tree produced by fodcatParser#expression.
    # @param ctx:fodcatParser.CommandContext
    def enterExpression(self, ctx: fodcatParser.ExpressionContext):
        self.expressionFlag = True  # Set the expression flag to true
        pass

    # Exit a parse tree produced by fodcatParser#expression.
    # @param ctx:fodcatParser.CommandContext
    def exitExpression(self, ctx: fodcatParser.ExpressionContext):
        self.expressionFlag = False  # Set the expression flag to false
        pass

    # ************************************************************************************
    # Declarations
    # ************************************************************************************

    # Enter a parse tree produced by fodcatParser#function_declaration.
    # @param ctx:fodcatParser.CommandContext
    def enterFunction_declaration(self, ctx: fodcatParser.Function_declarationContext):
        self.program += "global " + ctx.getChild(1).getText() + "\n"
        # Is function declaration without parameters
        if ctx.getChild(3).getChildCount() == 0:
            self.globals[ctx.getChild(1).getText(
            )] = "def " + ctx.getChild(1).getText() + "()"
            self.program += "def " + ctx.getChild(1).getText() + "():\n"
        else:  # Is function declaration with parameters
            self.globals[ctx.getChild(1).getText()] = "def " + ctx.getChild(1).getText() + \
                "(" + ctx.getChild(3).getText().replace('[]', '') + ")"
            self.program += "def " + \
                ctx.getChild(1).getText(
                ) + "(" + ctx.getChild(3).getText().replace('[]', '') + "):\n"
        self.localsFlag = True  # Set the local definitions flag to true
        self.countIndent += 1  # Add 1 to the indentation level
        if ctx.getChildCount() == 6:  # If the function body is empty
            self.indent(self.countIndent)  # Set proper indentation level
            self.program += "pass\n"  # Do nothing

    # Exit a parse tree produced by fodcatParser#function_declaration.
    # @param ctx:fodcatParser.CommandContext
    def exitFunction_declaration(self, ctx: fodcatParser.Function_declarationContext):
        self.countIndent = 0  # Set the indentation level to 0
        self.program += "\n"
        self.localsFlag = False  # Set the local definitions flag to false

    # Enter a parse tree produced by fodcatParser#array_declaration.
    # @param ctx:fodcatParser.CommandContext
    def enterArray_declaration(self, ctx: fodcatParser.Array_declarationContext):
        self.indent(self.countIndent)  # Set proper indentation level
        valueFlag = False  # Value is an expression
        if self.localsFlag:  # If is a local array declaration
            value = ctx.getChild(3).getText()
            if value in self.locals:
                value = int(self.locals[value])
            elif value in self.globals:
                value = int(self.globals[value])
            else:
                valueFlag = True  # Value is an expression
            try:
                # Checks if the value is an integer number
                self.locals[ctx.getChild(0).getText()] = [None] * int(value)
            except:
                if(not valueFlag):  # Is an identifier
                    self.locals[ctx.getChild(0).getText()] = [None] * value
                else:  # Is an expression
                    self.locals[ctx.getChild(0).getText()] = [None]
                self.canExecuteFlag = False
            # self.locals[ctx.getChild(0).getText()] = [None] * int(ctx.getChild(3).getText())
        else:  # If is a global array declaration
            value = ctx.getChild(3).getText()
            if value in self.globals:
                value = int(self.globals[value])
            else:
                valueFlag = True  # Value is an expression
            try:
                # Checks if the value is an integer number
                self.globals[ctx.getChild(0).getText()] = [None] * int(value)
            except:  # Is an identifier
                if (not valueFlag):  # Is an identifier
                    self.globals[ctx.getChild(0).getText()] = [None] * value
                else:  # Is an expression
                    self.globals[ctx.getChild(0).getText()] = [None]
                self.canExecuteFlag = False
            # self.globals[ctx.getChild(0).getText()] = [None] * int(ctx.getChild(3).getText())
        self.program += ctx.getChild(0).getText() + \
            " = [None] * " + ctx.getChild(3).getText()

    # Exit a parse tree produced by fodcatParser#array_declaration.
    # @param ctx:fodcatParser.CommandContext
    def exitArray_declaration(self, ctx: fodcatParser.Array_declarationContext):
        self.program += "\n"

    # Enter a parse tree produced by fodcatParser#gpio_declaration.
    # @param ctx:fodcatParser.CommandContext
    def enterGpio_declaration(self, ctx: fodcatParser.Gpio_declarationContext):
        self.hasGPIOFlag = True
        if ctx.getChild(3).getText() == ".":  # Is a variable brightness LED declaration
            if not self.hasVLEDFlag:
                self.program = "from gpiozero import PWMLED\n" + \
                    self.program  # Import the GPIO class
                self.hasVLEDFlag = True  # Has a variable LED definition
            value = GTI.getTextInterval(ctx, ctx.getChild(6))
            value = self.processKeyword(value)  # Get translated value
            # Checks if the value is an expression
            expressionList = self.checkIfExpression(value)
            if (any(expressionList)):  # If value is an expression
                if not self.localsFlag:  # If is a global declaration or value assignation
                    self.canExecuteFlag = False  # Will not be executed
            else:
                try:
                    self.globals[ctx.getChild(0).getText()] = int(
                        value)  # Checks if the value is an integer number
                except:
                    pass
            self.program += ctx.getChild(0).getText() + \
                " = PWMLED(" + value + ")"

        elif ctx.getChild(4).getChildCount() == 5:  # Is an angular servo declaration
            if not self.hasAngularServoFlag:
                self.program = "from gpiozero import AngularServo\n" + \
                    self.program  # Import the GPIO class
                self.hasAngularServoFlag = True
            value = GTI.getTextInterval(ctx, ctx.getChild(4))
            value = self.processKeyword(value)  # Get translated value
            # expressionList = self.checkIfExpression(value) # Checks if the value is an expression
            value = value.split(",")  # Separate the parameters
            expressionList = list()
            i = 0
            while (i < 3):  # Check if is a negative number or a expression
                # Checks if the value is an expression
                expressionList.append(self.checkIfExpression(value[i]))
                i += 1
            # If value is an expression
            if (any(expressionList[0]) & any(expressionList[1]) & any(expressionList[2])):
                if not self.localsFlag:  # If is a global declaration or value assignation
                    self.canExecuteFlag = False  # Will not be executed
            else:
                try:
                    self.globals[ctx.getChild(0).getText()] = [int(value[0]), int(
                        value[1]), int(value[2])]  # Checks if the values are an integer number
                except:
                    pass
            self.program += ctx.getChild(0).getText() + " = AngularServo(" + ctx.getChild(4).getChild(0).getText(
            ) + ", min_angle=" + ctx.getChild(4).getChild(2).getText() + ", max_angle=" + ctx.getChild(4).getChild(4).getText() + ")"

        elif ctx.getChild(2).getText() == "boton":  # Is a button declaration
            if not self.hasButtonFlag:
                self.program = "from gpiozero import Button\n" + \
                    self.program  # Import the GPIO class
                self.hasButtonFlag = True
            value = GTI.getTextInterval(ctx, ctx.getChild(4))
            value = self.processKeyword(value)  # Get translated value
            # Checks if the value is an expression
            expressionList = self.checkIfExpression(value)
            if (any(expressionList)):  # If value is an expression
                if not self.localsFlag:  # If is a global declaration or value assignation
                    self.canExecuteFlag = False  # Will not be executed
            else:
                try:
                    self.globals[ctx.getChild(0).getText()] = int(
                        value)  # Checks if the value is an integer number
                except:
                    pass
            self.program += ctx.getChild(0).getText() + \
                " = Button(" + ctx.getChild(4).getText() + ")"

        elif ctx.getChild(2).getText() == "led":  # Is a LED declaration
            if not self.hasLEDFlag:
                self.program = "from gpiozero import LED\n" + \
                    self.program  # Import the GPIO class
                self.hasLEDFlag = True
            value = GTI.getTextInterval(ctx, ctx.getChild(4))
            value = self.processKeyword(value)  # Get translated value
            # Checks if the value is an expression
            expressionList = self.checkIfExpression(value)
            if (any(expressionList)):  # If value is an expression
                if not self.localsFlag:  # If is a global declaration or value assignation
                    self.canExecuteFlag = False  # Will not be executed
            else:
                try:
                    self.globals[ctx.getChild(0).getText()] = int(
                        value)  # Checks if the value is an integer number
                except:
                    pass
            self.program += ctx.getChild(0).getText() + \
                " = LED(" + ctx.getChild(4).getText() + ")"

        elif ctx.getChild(2).getText() == "servo":  # Is a servo declaration
            if not self.hasServoFlag:
                self.program = "from gpiozero import Servo\n" + \
                    self.program  # Import the GPIO class
                self.hasServoFlag = True
            value = GTI.getTextInterval(ctx, ctx.getChild(4))
            value = self.processKeyword(value)  # Get translated value
            # Checks if the value is an expression
            expressionList = self.checkIfExpression(value)
            if (any(expressionList)):  # If value is an expression
                if not self.localsFlag:  # If is a global declaration or value assignation
                    self.canExecuteFlag = False  # Will not be executed
            else:
                try:
                    self.globals[ctx.getChild(0).getText()] = int(
                        value)  # Checks if the value is an integer number
                except:
                    pass
            self.program += ctx.getChild(0).getText() + \
                " = Servo(" + ctx.getChild(4).getText() + ")"

    # Exit a parse tree produced by fodcatParser#gpio_declaration.
    # @param ctx:fodcatParser.CommandContext
    def exitGpio_declaration(self, ctx: fodcatParser.Gpio_declarationContext):
        self.program += "\n"


# Verify the device where the compiler is been executed
# @return Boolean value if is running over a Raspberry Pi
def verifyDevice():
    if platform.system() == "Linux":  # If is a Linux OS
        command = "\n cat /proc/cpuinfo"
        all_info = subprocess.check_output(command, shell=True).strip()
        all_info = all_info.decode("utf-8")
        for line in all_info.split("\n"):
            if "model name" in line:
                # If is running over an ARM processor (Raspberry Pi)
                if "ARM" in re.sub(".*model name.*:", "", line, 1):
                    return True
    return False


def main(argv):
    # if (len(argv) == 4) & (argv[3] != "-f"):
    #     print("Error, la bandera para producir un archivo es -f.")
    #     quit()

    canGPIOFlag = verifyDevice()
    inp = FileStream(argv[1])
    lexer = fodcatLexer(inp)
    stream = CommonTokenStream(lexer)
    parser = fodcatParser(stream)
    tree = parser.program()

    if (len(argv) == 3):
        # if (len(argv) == 4):
        outfile = argv[2] + ".py"
        output = open(str(outfile), "w")
        fcat = CodeGenerator(output, canGPIOFlag)
    # elif (len(argv) == 1):
    #     output = None
    #     fcat = CodeGenerator(output, canGPIOFlag)
    else:
        print("Error, instruccion incorrecta.")
        quit()

    walker = ParseTreeWalker()
    walker.walk(fcat, tree)
    if (len(argv) == 4):
        output.close()


if __name__ == '__main__':
    main(sys.argv)



####################################################
# Miscellaneous Functions - LIE++ Language compiler
# APC-FOD
# Authors:
#   Ellioth Ramirez
####################################################
import subprocess
import sys
import json
from gen.fodcatListener import fodcatListener


# Documentation for Miscellaneous Functions

class multiFun(fodcatListener):
    # This function is used to fixed the problem with the getText,
    # because this one can get an interval, but it can't be a touple,
    # it's to be two separated variables.
    def getTextInterval(self, ctxParent, ctxSon):
        interval = ctxSon.getSourceInterval()
        return ctxParent.parser.getTokenStream().getText(interval[0], interval[1])

    # This function is used to execute the code from the memory.
    # this one can also return error in the process of execution
    # def execCode(self, code, args):
    def execCode(self, code):
        command = ['python', '-c', code]
        # command.extend(args)
        runfile = subprocess.run(command, stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        #stdout, stderr = runfile.communicate()
        if runfile.returncode != 0:
            return runfile.stderr.decode("utf-8").split("\r\n")[:-1]
        return runfile.stdout.decode("utf-8").split("\r\n")[:-1]

    # function to print a list of data.
    def printPreatty(self, stringList):
        for i in stringList:
            print(i)

    # this function it's done since we need to replace the characters that are change
    # because of the java code
    def replaceBadCharacters(self, code):
        code.replace("\'", " \"")
        #code.replace("\\n", " \n")
        return code



####################################################
# LIE++ Semantic Errors
# Definitions of all kinds of semantic errors
# handler by the compiler
# TEC - FOD
# Authors:
#   Michael Gonzalez Rivera
#   Luis Saborio Hernandez
####################################################

## Semantic errors base clase. Data class.
class SemanticError():
    
    ## Constructor method
    # @param line line number where the error ocurred.
    # @param column Character position where the error ocurred.
    def __init__(self, line:int, column:int):
        self.line = line
        self.column = column

## Used when there is an array call to an identifier wich is no being defined
class ArrayNotDefined(SemanticError):
    
    ## Constructor method
    # @param line line number where the error ocurred.
    # @param column Character position where the error ocurred.
    # @param offended_identifier identifier wich is no beeing defined
    def __init__(self, line:int, column:int, offended_identifier:str):
        super().__init__(line, column)
        self.offended_identifier = offended_identifier

## Used when there is a call to an identifier defined as array, bu it do not specify an index.
class ArrayWithoutIndex(SemanticError):
    
    ## Constructor method
    # @param line line number where the error ocurred.
    # @param column Character position where the error ocurred.
    # @param offended_identifier identifier wich is no beeing defined
    def __init__(self, line:int, column:int, offended_identifier:str):
        super().__init__(line, column)
        self.offended_identifier = offended_identifier

## Used when a variable is redefined. The only two types of variables are arrays and scalar variables.
class VariableRedefinition(SemanticError):
    
    ## Constructor method
    # @param line line number where the error ocurred.
    # @param column Character position where the error ocurred.
    # @param duplicated_identifier identifier wich is beeing redefined
    def __init__(self, line:int, column:int, duplicated_identifier:str):
        super().__init__(line, column)
        self.duplicated_identifier = duplicated_identifier
        
## Used when there is a redefinition function.
# The function signature is the name and the quantity of parameters.
class FunctionRedefinition(SemanticError):
    
    ## Constructor method
    # @param line line number where the error ocurred.
    # @param column Character position where the error ocurred.
    # @param duplicated_identifier function name wich is beeing redefined.
    # @param params_qty function parameters quantity.
    def __init__(self, line:int, column:int, duplicated_identifier:str, params_qty:int):
        super().__init__(line, column)
        self.duplicated_identifier = duplicated_identifier
        self.params_qty = params_qty

## Used when there is function call to a function signature not defined before.
class NoMatchedFunction(SemanticError):

    ## Constructor method
    # @param line line number where the error ocurred.
    # @param column Character position where the error ocurred.
    # @param duplicated_identifier function name wich is beeing redefined.
    # @param params_qty function parameters quantity.
    def __init__(self, line:int, column:int, no_matched_identifier:str, params_qty:int):
        super().__init__(line, column)
        self.no_matched_identifier = no_matched_identifier
        self.params_qty = params_qty

## Used when there is a call to a identifier wich is not beeing defined before.
class NoMatchIdentifier(SemanticError):
    
    ## Constructor method
    # @param line line number where the error ocurred.
    # @param column Character position where the error ocurred.
    # @param no_matched_identifier idenfier wich is not beeing deifned.
    def __init__(self, line:int, column:int, no_matched_identifier:str):
        super().__init__(line, column)
        self.no_matched_identifier = no_matched_identifier

## Used when sending a parameter wich do not match with the parameter type of the function signature.
class NoCompatibleParam(SemanticError):

    ## Constructor method
    # @param line line number where the error ocurred.
    # @param column Character position where the error ocurred.
    # @param expected_type Expected parameter type    # @param getted_type Sended parameter type.
    def __init__(self, line:int, column:int, expected_type:str, getted_type:str):
        super().__init__(line, column)
        self.expected_type = expected_type
        self.getted_type = getted_type


####################################################
# Code analyzer to LIE + + Language
# TEC - FOD
# Authors:
#   Michael Gonzalez Rivera
#   Luis Saborio Hernandez
# Edited:
#   Ellioth Ramirez
####################################################

import sys
import json

from antlr4 import *
from gen.fodcatListener import fodcatListener
from gen.fodcatLexer import fodcatLexer
from gen.fodcatParser import fodcatParser

from multi.multiFun import multiFun
GTI = multiFun()

# Documentation for CodeAnalyzer
#
# Code analyzer that extracts information on the number of times each type of instruction
# is used, as well as the definitions of the functions, expressions, operators, and
# reserved words used, among others.


class CodeAnalyzer(fodcatListener):

    # CodeAnalyzer constructor. Initialize all needed global variables for the
    # succesfully execution of the code analyzer.
    # @param output JSON analysis output file name.
    def __init__(self, output):
        # Store the Python output file name.
        self.output = output

        # Flag to store if it is an array
        self.arrayFlag = False
        # Flag to store if it is a function declaration
        self.functionDecFlag = False
        # Flag to store if it is a function declaration identifier
        self.functionIdenDecFlag = False
        # Flag to store if it is a function call
        self.functionCallFlag = False
        # Flag to store if it is a cycle
        self.cycleFlag = False
        # Flag to store if it is a conditional
        self.conditionalFlag = False
        # Flag to store if it is the main function
        self.mainFlag = False

        # Dictionary to store all the rules and subrules during the analysis
        self.ruleDictionary = {}
        self.ruleDictionary["variables"] = []
        self.ruleDictionary["arreglos"] = []
        self.ruleDictionary["sentencias"] = []
        self.ruleDictionary["identificadores"] = []
        self.ruleDictionary["enteros"] = []
        self.ruleDictionary["flotantes"] = []
        self.ruleDictionary["booleanos"] = []
        self.ruleDictionary["texto"] = []
        self.ruleDictionary["constantes"] = []
        self.ruleDictionary["expresion aritmetica"] = []
        self.ruleDictionary["operadores aritmeticos"] = []
        self.ruleDictionary["expresion logica"] = []
        self.ruleDictionary["operadores logicos"] = []
        self.ruleDictionary["expresion relacional"] = []
        self.ruleDictionary["operadores relacionales"] = []
        self.ruleDictionary["signos"] = []
        self.ruleDictionary["control flujo"] = []
        self.ruleDictionary["funcion principal"] = []
        # Dictionary to store all the keywords during the analysis
        self.keywords = {}
        self.keywords["palabras reservadas"] = []
        # Dictionary to count rules, subrules and tokens
        self.rulecount = {
            "cantidad variables": 0,
            "cantidad arreglos": 0,
            "cantidad funciones": 0,
            "cantidad sentencias": 0,
            "llamadas variables": 0,
            "llamadas arreglos": 0,
            "llamadas funciones": 0,
            "cantidad expresiones aritmeticas": 0,
            "cantidad expresiones logicas": 0,
            "cantidad expresiones relacionales": 0,
            "operador aritmetico": 0,
            "operador logico": 0,
            "operador relacional": 0,
            "cantidad booleanos": 0,
            "cantidad enteros": 0,
            "cantidad flotantes": 0,
            "cantidad texto": 0,
            "cantidad constantes matematicas": 0,
            "cantidad signos": 0,
            "si": 0,
            "sino": 0,
            "anidacion condicionales": 0,
            "mientras": 0,
            "cantidad control flujo": 0,
            "devolver": 0,
            "repetir": 0,
            "imprimir": 0,
            "operador imprimir": 0,
            "operador asignacion": 0,
            "cantidad identificadores": 0,
            "anidacion ciclos": 0,
        }

        # List to store variable names
        self.varNames = []
        # List to store everything within the function declaration
        self.funDecDic = []
        # Stores the function declaration
        self.funDec = ""

        # Stores the nesting level of the conditionals
        self.conditionalNesting = 0
        # Stores the nesting level of the cycles
        self.cycleNesting = 0

    # Method to fill the keyword dictionary. Adds keyword of list into dictionary of key
    # "palabras reservadas".
    # @param keyWord List of keywords.
    def addKeywords(self, keyWord):
        if (not self.keywords["palabras reservadas"].__contains__(keyWord)):
            self.keywords["palabras reservadas"].append(keyWord)

    # Method to write multiple dictionary to JSON file, separated by comma. Writes the
    # key separated by : and writes de value of the dictionary compares last rule to not
    # write the last comma
    # @param ctx:fodcatParser.dictionary Any python dictionary
    def printDictionary(self, dictionary):
        for x, y in dictionary.items():
            self.output.write("\n")
            if (str(x) == "palabras reservadas"):  # Compare last rule of file to not write a comma
                self.output.write('"' + str(x) + '"' +
                                  ":" + str(json.dumps(y)))
            else:
                self.output.write('"' + str(x) + '"' +
                                  ":" + str(json.dumps(y)))
                self.output.write(", ")

    # ************************************************************************************
    # Initial Rule
    # ************************************************************************************

    # Exit a parse tree produced by fodcatParser#program. Writes in a file the analysis
    # result in JSON.
    # @param ctx:fodcatParser.ProgramContext
    def exitProgram(self, ctx: fodcatParser.ProgramContext):
        self.output.write("[")
        self.output.write("{")
        self.printDictionary(self.rulecount)
        self.printDictionary(self.ruleDictionary)
        writeFunDec = {}
        writeFunDec["funciones"] = self.funDecDic
        self.printDictionary(writeFunDec)
        self.printDictionary(self.keywords)
        self.output.write("}")
        self.output.write("]")

    # Enter a parse tree produced by fodcatParser#main_function.
    # @param ctx:fodcatParser.Main_functionContext
    def enterMain_function(self, ctx: fodcatParser.Main_functionContext):
        self.mainFlag = True  # It is the main function
        self.ruleDictionary["funcion principal"].append("main")

    # Exit a parse tree produced by fodcatParser#main_function.
    # @param ctx:fodcatParser.Main_functionContext
    def exitMain_function(self, ctx: fodcatParser.Main_functionContext):
        self.mainFlag = False  # It is no longer in the main function
        self.addKeywords("para")
        self.addKeywords("principal()")
        self.addKeywords("fin")

    # ************************************************************************************
    # Commands
    # ************************************************************************************

    # Enter a parse tree produced by fodcatParser#command.
    # @param ctx:fodcatParser.CommandContext
    def enterCommand(self, ctx: fodcatParser.CommandContext):
        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # If the command is a "mientras" statement
        if ctx.getChild(0).getText() == "mientras":
            self.rulecount["mientras"] += 1  # Add 1 to the counter
            self.rulecount["cantidad sentencias"] += 1  # Add 1 to the counter

            if (self.cycleFlag):
                self.cycleNesting += 1  # Add 1 to the counter
            self.cycleFlag = True  # Is a cycle

            # Check if dictionary "sentencias" has "mientras", if not, add
            if (not self.ruleDictionary["sentencias"].__contains__("mientras")):
                self.ruleDictionary["sentencias"].append("mientras")

            whileStat = ""
            j = 0
            while(j < ctx.getChildCount()):
                if (not ctx.getChild(j).getText() == "haga"):
                    # Creates the whileStat
                    whileStat += ctx.getChild(j).getText() + " "
                else:
                    j = ctx.getChildCount()
                j += 1  # Add 1 to the counter
            whileStat = whileStat[:-1]  # Deletes the last character (space)
            # Checks where it has to add the information
            if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
                if (not self.funDecDic[i]["cuerpo funcion"].__contains__("sentencia: " + whileStat)):
                    self.funDecDic[i]["cuerpo funcion"].append(
                        "sentencia: " + whileStat)

            if (self.functionDecFlag and self.cycleFlag):
                if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("sentencia: " + whileStat)):
                    self.funDecDic[i]["cuerpo ciclo"].append(
                        "sentencia: " + whileStat)

            if (self.functionDecFlag and self.conditionalFlag):
                if (not self.funDecDic[i]["cuerpo condicional"].__contains__("sentencia: " + whileStat)):
                    self.funDecDic[i]["cuerpo condicional"].append(
                        "sentencia: " + whileStat)

        # If the command is a "repetir" statement
        elif ctx.getChild(0).getText() == "repetir":
            self.rulecount["repetir"] += 1  # Add 1 to the counter
            self.rulecount["cantidad sentencias"] += 1  # Add 1 to the counter

            if (self.cycleFlag):
                self.cycleNesting += 1  # Add 1 to the counter
            self.cycleFlag = True

            if (not self.ruleDictionary["sentencias"].__contains__("repetir")):
                self.ruleDictionary["sentencias"].append(
                    "repetir")  # Adds "repetir" to the dictionary

            repeatStat = ""
            j = 0
            while(j < ctx.getChildCount()):
                if (not ctx.getChild(j).getText() == "veces"):
                    # Creates the repeatStat
                    repeatStat += ctx.getChild(j).getText() + " "
                else:
                    j = ctx.getChildCount()
                j += 1  # Add 1 to the counter
            repeatStat = repeatStat[:-1]  # Deletes the last character (space)

            # Checks where it has to add the information
            if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
                if (not self.funDecDic[i]["cuerpo funcion"].__contains__("sentencia: " + repeatStat)):
                    self.funDecDic[i]["cuerpo funcion"].append(
                        "sentencia: " + repeatStat)

            if (self.functionDecFlag and self.cycleFlag):
                if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("sentencia: " + repeatStat)):
                    self.funDecDic[i]["cuerpo ciclo"].append(
                        "sentencia: " + repeatStat)

            if (self.functionDecFlag and self.conditionalFlag):
                if (not self.funDecDic[i]["cuerpo condicional"].__contains__("sentencia: " + repeatStat)):
                    self.funDecDic[i]["cuerpo condicional"].append(
                        "sentencia: " + repeatStat)

        # If the command is a "imprimir" statement
        elif ctx.getChild(0).getText() == "imprimir":
            self.rulecount["imprimir"] += 1  # Add 1 to the counter
            self.rulecount["cantidad sentencias"] += 1  # Add 1 to the counter
            if (self.mainFlag):  # Is in the main function
                # Adds "imprimir" to the dictionary
                value = GTI.getTextInterval(ctx, ctx)
                self.ruleDictionary["funcion principal"].append(
                    "sentencia imprimir: " + value)
            else:
                if (not self.ruleDictionary["sentencias"].__contains__("imprimir")):
                    self.ruleDictionary["sentencias"].append(
                        "imprimir")  # Adds "imprimir" to the dictionary

            printStat = ""
            j = 0
            while(j < ctx.getChildCount()):
                # Creates the printStat
                printStat += ctx.getChild(j).getText() + " "
                j += 1  # Add 1 to the counter
            printStat = printStat[:-1]  # Deletes the last character (space)

            # Checks where it has to add the information
            if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
                if (not self.funDecDic[i]["cuerpo funcion"].__contains__("sentencia: " + printStat)):
                    self.funDecDic[i]["cuerpo funcion"].append(
                        "sentencia: " + printStat)

            if (self.functionDecFlag and self.cycleFlag):
                if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("sentencia: " + printStat)):
                    self.funDecDic[i]["cuerpo ciclo"].append(
                        "sentencia: " + printStat)

            if (self.functionDecFlag and self.conditionalFlag):
                if (not self.funDecDic[i]["cuerpo condicional"].__contains__("sentencia: " + printStat)):
                    self.funDecDic[i]["cuerpo condicional"].append(
                        "sentencia: " + printStat)

        # If the command is a "devolver" statement
        elif ctx.getChild(0).getText() == "devolver":
            self.rulecount["devolver"] += 1  # Add 1 to the counter
            self.rulecount["cantidad sentencias"] += 1  # Add 1 to the counter

            if (not self.ruleDictionary["sentencias"].__contains__("devolver")):
                self.ruleDictionary["sentencias"].append(
                    "devolver")  # Adds "devolver" to the dictionary

            returnStat = ""
            j = 0
            while(j < ctx.getChildCount()):
                # Creates the returnStat
                returnStat += ctx.getChild(j).getText() + " "
                j += 1  # Add 1 to the counter
            returnStat = returnStat[:-1]  # Deletes the last character (space)

            # Checks where it has to add the information
            if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
                if (not self.funDecDic[i]["cuerpo funcion"].__contains__("sentencia: " + returnStat)):
                    self.funDecDic[i]["cuerpo funcion"].append(
                        "sentencia: " + returnStat)

            if (self.functionDecFlag and self.cycleFlag):
                if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("sentencia: " + returnStat)):
                    self.funDecDic[i]["cuerpo ciclo"].append(
                        "sentencia: " + returnStat)

            if (self.functionDecFlag and self.conditionalFlag):
                if (not self.funDecDic[i]["cuerpo condicional"].__contains__("sentencia: " + returnStat)):
                    self.funDecDic[i]["cuerpo condicional"].append(
                        "sentencia: " + returnStat)
        else:
            pass

    # Exit a parse tree produced by fodcatParser#command.
    # @param ctx:fodcatParser.CommandContext
    def exitCommand(self, ctx: fodcatParser.CommandContext):
        # If the command is a "mientras" statement
        if ctx.getChild(0).getText() == "mientras":
            self.addKeywords("mientras")
            self.addKeywords("haga")
            self.addKeywords("fin")
            if self.cycleNesting > self.rulecount["anidacion ciclos"]:
                self.rulecount["anidacion ciclos"] = self.cycleNesting
            self.cycleNesting = 0
            self.cycleFlag = False
        # If the command is a "repetir" statement
        elif ctx.getChild(0).getText() == "repetir":
            self.addKeywords("repetir")
            self.addKeywords("veces")
            self.addKeywords("fin")
            if self.cycleNesting > self.rulecount["anidacion ciclos"]:
                self.rulecount["anidacion ciclos"] = self.cycleNesting
            self.cycleNesting = 0
            self.cycleFlag = False
        # If the command is a "imprimir" statement
        elif ctx.getChild(0).getText() == "imprimir":
            self.addKeywords("imprimir")
        # If the command is a "devolver" statement
        elif ctx.getChild(0).getText() == "devolver":
            self.addKeywords("devolver")
        else:
            pass

    # Enter a parse tree produced by fodcatParser#ifstat.
    # @param ctx:fodcatParser.IfstatContext
    def enterIfstat(self, ctx: fodcatParser.IfstatContext):
        self.rulecount["si"] += 1  # Add 1 to the counter
        self.rulecount["cantidad sentencias"] += 1  # Add 1 to the counter

        if(self.conditionalFlag):  # Count levels of nested conditionals
            self.conditionalNesting += 1  # Add 1 to the counter

        self.conditionalFlag = True  # Is a conditional

        # Check if "si" is the current dictionary with key "sentencias", if not, add
        if (not self.ruleDictionary["sentencias"].__contains__("si")):
            self.ruleDictionary["sentencias"].append(
                "si")  # Adds "si" to the dictionary

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        ifStat = GTI.getTextInterval(ctx, ctx)
        ifStat = ifStat.split(" entonces")[0]  # Creates the ifStat

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("sentencia: " + ifStat)):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "sentencia: " + ifStat)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("sentencia: " + ifStat)):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "sentencia: " + ifStat)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("sentencia: " + ifStat)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "sentencia: " + ifStat)

    # Exit a parse tree produced by fodcatParser#ifstat.
    # @param ctx:fodcatParser.IfstatContext
    def exitIfstat(self, ctx: fodcatParser.IfstatContext):
        self.addKeywords("si")
        self.addKeywords("entonces")
        self.addKeywords("fin")
        if self.conditionalNesting > self.rulecount["anidacion condicionales"]:
            self.rulecount["anidacion condicionales"] = self.conditionalNesting
        self.conditionalNesting = 0
        self.conditionalFlag = False

    # Enter a parse tree produced by fodcatParser#elsestat.
    # @param ctx:fodcatParser.ElsestatContext
    def enterElsestat(self, ctx: fodcatParser.ElsestatContext):
        self.rulecount["sino"] += 1  # Add 1 to the counter
        self.rulecount["cantidad sentencias"] += 1  # Add 1 to the counter

        # Check if "si" is the current dictionary with key "sentencias", if not, add
        if (not self.ruleDictionary["sentencias"].__contains__("sino")):
            self.ruleDictionary["sentencias"].append(
                "sino")  # Adds "sino" to the dictionary

    # Exit a parse tree produced by fodcatParser#elsestat.
    # @param ctx:fodcatParser.ElsestatContext
    def exitElsestat(self, ctx: fodcatParser.ElsestatContext):
        self.addKeywords("sino")

    # Enter a parse tree produced by fodcatParser#func_call.
    # @param ctx:fodcatParser.Func_callContext
    def enterFunc_call(self, ctx: fodcatParser.Func_callContext):
        self.rulecount["llamadas funciones"] += 1  # Add 1 to the counter
        self.rulecount["cantidad sentencias"] += 1  # Add 1 to the counter
        self.functionCallFlag = True
        value = GTI.getTextInterval(ctx, ctx)
        if (self.mainFlag):  # Is in the main function
            self.ruleDictionary["funcion principal"].append(
                "llamada funcion: " + value)
        else:
            if (not self.ruleDictionary["sentencias"].__contains__("sentencia funcion")):
                self.ruleDictionary["sentencias"].append(
                    "sentencia funcion")  # Adds "sentencia" to the dictionary

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("llamada funcion: " + value)):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "llamada funcion: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("llamada funcion: " + value)):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "llamada funcion: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("llamada funcion: " + value)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "llamada funcion: " + value)

    # Exit a parse tree produced by fodcatParser#func_call.
    # @param ctx:fodcatParser.Func_callContext
    def exitFunc_call(self, ctx: fodcatParser.Func_callContext):
        self.functionCallFlag = False

    # Enter a parse tree produced by fodcatParser#flow_exp.
    # @param ctx:fodcatParser.Flow_expContext
    def enterFlow_exp(self, ctx: fodcatParser.Flow_expContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["cantidad control flujo"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["control flujo"].__contains__(value)):
            self.ruleDictionary["control flujo"].append(
                value)  # Adds "control flujo" to the dictionary

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("control flujo: " + value)):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "control flujo: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("control flujo: " + value)):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "control flujo: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("control flujo: " + value)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "control flujo: " + value)

    # Exit a parse tree produced by fodcatParser#flow_exp.
    # @param ctx:fodcatParser.Flow_expContext
    def exitFlow_exp(self, ctx: fodcatParser.Flow_expContext):
        self.addKeywords(ctx.getChild(0).getText())

    # Enter a parse tree produced by fodcatParser#assignment.
    # @param ctx:fodcatParser.AssignmentContext
    def enterAssignment(self, ctx: fodcatParser.AssignmentContext):
        self.rulecount["operador asignacion"] += 1  # Add 1 to the counter

    # ************************************************************************************
    # Expressions
    # ************************************************************************************

    # Enter a parse tree produced by fodcatParser#expression.
    # @param ctx:fodcatParser.ExpressionContext
    def enterExpression(self, ctx: fodcatParser.ExpressionContext):
        if ctx.getChild(0).getChildCount() > 1:
            if ctx.getChild(0).getChild(1).getText() == "[":
                self.arrayFlag = True  # Is an array

    # Exit a parse tree produced by fodcatParser#expression.
    # @param ctx:fodcatParser.ExpressionContext
    def exitExpression(self, ctx: fodcatParser.ExpressionContext):
        if ctx.getChild(0).getChildCount() > 1:
            if ctx.getChild(0).getChild(1).getText() == "[":
                self.arrayFlag = False  # Was an array

    # Enter a parse tree produced by fodcatParser#arithmetic_exp.
    # @param ctx:fodcatParser.
    def enterArithmetic_exp(self, ctx: fodcatParser.Arithmetic_expContext):
        value = GTI.getTextInterval(ctx, ctx)
        # Add 1 to the counter
        self.rulecount["cantidad expresiones aritmeticas"] += 1
        if (not self.ruleDictionary["expresion aritmetica"].__contains__(value)):
            # Adds "expresion aritmetica" to the dictionary
            self.ruleDictionary["expresion aritmetica"].append(value)
        if ctx.getChild(0).getChildCount() > 1:
            if ctx.getChild(0).getChild(1).getText() == "[":
                self.arrayFlag = True  # Is an array

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("expresion aritmetica: " + value)):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "expresion aritmetica: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("expresion aritmetica: " + value)):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "expresion aritmetica: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("expresion aritmetica: " + value)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "expresion aritmetica: " + value)

    # Enter a parse tree produced by fodcatParser#relational_exp.
    # @param ctx:fodcatParser.
    def enterRelational_exp(self, ctx: fodcatParser.Relational_expContext):
        value = GTI.getTextInterval(ctx, ctx)
        # Add 1 to the counter
        self.rulecount["cantidad expresiones relacionales"] += 1
        if not (self.ruleDictionary["expresion relacional"].__contains__(value) |
                self.ruleDictionary["expresion relacional"].__contains__("(" + value + ")")):
            # Adds "expresion relacional" to the dictionary
            self.ruleDictionary["expresion relacional"].append(value)

        if (self.ruleDictionary["expresion logica"].__contains__(value) |
                self.ruleDictionary["expresion logica"].__contains__("(" + value + ")")):
            # Adds "expresion logica" to the dictionary
            self.ruleDictionary["expresion logica"].pop(
                self.ruleDictionary["expresion logica"].index(value))

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if not (self.funDecDic[i]["cuerpo funcion"].__contains__("expresion relacional: " + value) |
                    self.funDecDic[i]["cuerpo funcion"].__contains__("expresion relacional: " + "(" + value + ")")):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "expresion relacional: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if not (self.funDecDic[i]["cuerpo ciclo"].__contains__("expresion relacional: " + value) |
                    self.funDecDic[i]["cuerpo ciclo"].__contains__("expresion relacional: " + "(" + value + ")")):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "expresion relacional: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if not (self.funDecDic[i]["cuerpo condicional"].__contains__("expresion relacional: " + value) |
                    self.funDecDic[i]["cuerpo condicional"].__contains__("expresion relacional: " + "(" + value + ")")):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "expresion relacional: " + value)

    # Enter a parse tree produced by fodcatParser#logic_exp.
    # @param ctx:fodcatParser.
    def enterLogic_exp(self, ctx: fodcatParser.Logic_expContext):
        value = GTI.getTextInterval(ctx, ctx)
        # Add 1 to the counter
        self.rulecount["cantidad expresiones logicas"] += 1
        if not (self.ruleDictionary["expresion relacional"].__contains__(value) |
                self.ruleDictionary["expresion relacional"].__contains__("(" + value + ")")):
            if not (self.ruleDictionary["expresion logica"].__contains__(value) |
                    self.ruleDictionary["expresion logica"].__contains__("(" + value + ")")):
                self.ruleDictionary["expresion logica"].append(
                    value)  # Adds "expresion logica" to the dictionary

        if ((ctx.getChild(0).getText() == "no") | (ctx.getChild(0).getText() == "|")):
            self.rulecount["operador logico"] += 1  # Add 1 to the counter
            if (not self.ruleDictionary["operadores logicos"].__contains__(ctx.getChild(0).getText())):
                self.ruleDictionary["operadores logicos"].append(ctx.getChild(
                    0).getText())  # Adds "operadores logicos" to the dictionary

        if ctx.getChild(0).getChildCount() > 1:
            if ctx.getChild(0).getChild(1).getText() == "[":
                self.arrayFlag = True  # Is an array

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if not (self.funDecDic[i]["cuerpo funcion"].__contains__("expresion logica: " + value) |
                    self.funDecDic[i]["cuerpo condicional"].__contains__("expresion funcion: " + "(" + value + ")")):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "expresion logica: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if not (self.funDecDic[i]["cuerpo ciclo"].__contains__("expresion logica: " + value) |
                    self.funDecDic[i]["cuerpo ciclo"].__contains__("expresion logica: " + "(" + value + ")")):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "expresion logica: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if not (self.funDecDic[i]["cuerpo condicional"].__contains__("expresion logica: " + value) |
                    self.funDecDic[i]["cuerpo condicional"].__contains__("expresion logica: " + "(" + value + ")")):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "expresion logica: " + value)

    def enterArray_call(self, ctx: fodcatParser.Array_callContext):
        self.arrayFlag = True  # Is an array

    # ************************************************************************************
    # Declarations
    # ************************************************************************************

    # Enter a parse tree produced by fodcatParser#function_declaration.
    # @param ctx:fodcatParser.Function_declarationContext
    def enterFunction_declaration(self, ctx: fodcatParser.Function_declarationContext):
        self.functionDecFlag = True  # Is a function declaration
        self.functionIdenDecFlag = True
        self.rulecount["cantidad funciones"] += 1  # Add 1 to the counter
        i = 1

        while (i < ctx.getChildCount()):
            self.funDec += ctx.getChild(i).getText()  # Creates the funDec
            if (ctx.getChild(i).getText() == ")"):
                i = ctx.getChildCount()
            else:
                i += 1  # Add 1 to the counter
        p = 0
        newFunDec = {
            "declaracion": self.funDec,
            "cuerpo funcion": [],
            "cuerpo ciclo": [],
            "cuerpo condicional": [],
        }
        self.funDecDic.append(newFunDec)  # Adds the newFunDec to the funDecDic

        # i=0
        # while (i<len(self.funDecDic)):
        #     print(self.funDecDic[i])
        #     print(self.funDecDic[i]["declaracion"].__contains__(self.funDec))
        #     i +=1

    # Exit a parse tree produced by fodcatParser#function_declaration.
    # @param ctx:fodcatParser.Function_declarationContext
    def exitFunction_declaration(self, ctx: fodcatParser.Function_declarationContext):
        self.functionDecFlag = False
        self.funDec = ""
        self.addKeywords("para")
        self.addKeywords("fin")

    # Enter a parse tree produced by fodcatParser#array_declaration.
    # @param ctx:fodcatParser.Array_declarationContext
    def enterArray_declaration(self, ctx: fodcatParser.Array_declarationContext):
        self.arrayFlag = True

    # Exit a parse tree produced by fodcatParser#array_declaration.
    # @param ctx:fodcatParser.Array_declarationContext
    def exitArray_declaration(self, ctx: fodcatParser.Array_declarationContext):
        self.arrayFlag = False

    # ************************************************************************************
    # Literals
    # ************************************************************************************

    # Enter a parse tree produced by fodcatParser#int_literal.
    # @param ctx:fodcatParser.
    def enterInt_literal(self, ctx: fodcatParser.Int_literalContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["cantidad enteros"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["enteros"].__contains__(value)):
            self.ruleDictionary["enteros"].append(
                value)  # Adds "enteros" to the dictionary

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("entero literal: " + value)):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "entero literal: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("entero literal: " + value)):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "entero literal: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("entero literal: " + value)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "entero literal: " + value)

    # Enter a parse tree produced by fodcatParser#bool_literal.
    # @param ctx:fodcatParser.
    def enterBool_literal(self, ctx: fodcatParser.Bool_literalContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["cantidad booleanos"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["booleanos"].__contains__(value)):
            self.ruleDictionary["booleanos"].append(
                value)  # Adds "booleanos" to the dictionary

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("booleano literal: " + value)):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "booleano literal: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("booleano literal: " + value)):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "booleano literal: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("booleano literal: " + value)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "booleano literal: " + value)

    # Enter a parse tree produced by fodcatParser#float_literal.
    # @param ctx:fodcatParser.
    def enterFloat_literal(self, ctx: fodcatParser.Float_literalContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["cantidad flotantes"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["flotantes"].__contains__(value)):
            self.ruleDictionary["flotantes"].append(
                value)  # Adds "flotantes" to the dictionary

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("flotante literal: " + value)):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "flotante literal: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("flotante literal: " + value)):
                self.funDecDic[i]["cuerpo ciclo"].append(
                    "flotante literal: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("flotante literal: " + value)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "flotante literal: " + value)

    # Enter a parse tree produced by fodcatParser#strg.
    # @param ctx:fodcatParser.
    def enterStrg(self, ctx: fodcatParser.StrgContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["cantidad texto"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["texto"].__contains__(value)):
            self.ruleDictionary["texto"].append(
                value)  # Adds "texto" to the dictionary

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("texto: " + value)):
                self.funDecDic[i]["cuerpo funcion"].append("texto: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("texto: " + value)):
                self.funDecDic[i]["cuerpo ciclo"].append("texto: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("texto: " + value)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "texto: " + value)

    # Enter a parse tree produced by fodcatParser#constant.
    # @param ctx:fodcatParser.
    def enterConstant(self, ctx: fodcatParser.ConstantContext):
        value = GTI.getTextInterval(ctx, ctx)
        # Add 1 to the counter
        self.rulecount["cantidad constantes matematicas"] += 1
        if (not self.ruleDictionary["constantes"].__contains__(value)):
            self.ruleDictionary["constantes"].append(
                value)  # Adds "constantes" to the dictionary

        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        # Checks where it has to add the information
        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("constante: " + value)):
                self.funDecDic[i]["cuerpo funcion"].append(
                    "constante: " + value)

        if (self.functionDecFlag and self.cycleFlag):
            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("constante: " + value)):
                self.funDecDic[i]["cuerpo ciclo"].append("constante: " + value)

        if (self.functionDecFlag and self.conditionalFlag):
            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("constante: " + value)):
                self.funDecDic[i]["cuerpo condicional"].append(
                    "constante: " + value)

    # Enter a parse tree produced by fodcatParser#relational_op.
    # @param ctx:fodcatParser.
    def enterRelational_op(self, ctx: fodcatParser.Relational_opContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["operador relacional"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["operadores relacionales"].__contains__(value)):
            # Adds "operadores relacionales" to the dictionary
            self.ruleDictionary["operadores relacionales"].append(value)

    # Enter a parse tree produced by fodcatParser#logic_op.
    # @param ctx:fodcatParser.
    def enterLogic_op(self, ctx: fodcatParser.Logic_opContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["operador logico"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["operadores logicos"].__contains__(value)):
            # Adds "operadores logicos" to the dictionary
            self.ruleDictionary["operadores logicos"].append(value)

    # Enter a parse tree produced by fodcatParser#arithmetic_op.
    # @param ctx:fodcatParser.
    def enterArithmetic_op(self, ctx: fodcatParser.Arithmetic_opContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["operador aritmetico"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["operadores aritmeticos"].__contains__(value)):
            # Adds "operadores aritmeticos" to the dictionary
            self.ruleDictionary["operadores aritmeticos"].append(value)

    # Enter a parse tree produced by fodcatParser#print_op.
    # @param ctx:fodcatParser.
    def enterPrint_op(self, ctx: fodcatParser.Print_opContext):
        self.rulecount["operador imprimir"] += 1  # Add 1 to the counter

    # Enter a parse tree produced by fodcatParser#sign.
    # @param ctx:fodcatParser.
    def enterSign(self, ctx: fodcatParser.SignContext):
        value = GTI.getTextInterval(ctx, ctx)
        self.rulecount["cantidad signos"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["signos"].__contains__(value)):
            # Adds "signos" to the dictionary
            self.ruleDictionary["signos"].append(value)

    # Enter a parse tree produced by fodcatParser#identifier.
    # @param ctx:fodcatParser.IdentifierContext
    def enterIdentifier(self, ctx: fodcatParser.IdentifierContext):
        identValue = GTI.getTextInterval(ctx, ctx)
        i = 0
        while (i < len(self.funDecDic)):
            # Checks if the function was already declared
            if (self.funDecDic[i]["declaracion"] == self.funDec):
                break
            i += 1  # Add 1 to the counter

        self.rulecount["cantidad identificadores"] += 1  # Add 1 to the counter
        if (not self.ruleDictionary["identificadores"].__contains__(identValue)):
            # Adds "identificadores" to the dictionary
            self.ruleDictionary["identificadores"].append(identValue)

        if (not self.functionCallFlag):  # If is not a function call
            if self.functionIdenDecFlag:
                self.functionIdenDecFlag = False
            else:
                if self.arrayFlag:  # Is an array
                    self.arrayFlag = False
                    if self.ruleDictionary["arreglos"].__contains__(identValue):
                        # Add 1 to the counter
                        self.rulecount["llamadas arreglos"] += 1
                    else:
                        # Add 1 to the counter
                        self.rulecount["cantidad arreglos"] += 1
                        self.ruleDictionary["arreglos"].append(
                            identValue)  # Adds the array to the dictionary
                    if (self.mainFlag):  # Is in the main function
                        self.ruleDictionary["funcion principal"].append(
                            " arreglo: " + identValue)  # Adds "arreglo" to the dictionary
                    else:
                        # Checks where it has to add the information
                        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
                            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("arreglo: " + identValue)):
                                self.funDecDic[i]["cuerpo funcion"].append(
                                    "arreglo: " + identValue)

                        if (self.functionDecFlag and self.cycleFlag):
                            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("arreglo: " + identValue)):
                                self.funDecDic[i]["cuerpo ciclo"].append(
                                    "arreglo: " + identValue)

                        if (self.functionDecFlag and self.conditionalFlag):
                            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("arreglo: " + identValue)):
                                self.funDecDic[i]["cuerpo condicional"].append(
                                    "arreglo: " + identValue)
                else:  # Is a variable
                    if self.ruleDictionary["variables"].__contains__(identValue):
                        # Add 1 to the counter
                        self.rulecount["llamadas variables"] += 1
                    else:
                        # Add 1 to the counter
                        self.rulecount["cantidad variables"] += 1
                        self.ruleDictionary["variables"].append(
                            identValue)  # Adds "variables" to the dictionary
                    if (self.mainFlag):  # Is in the main function
                        self.ruleDictionary["funcion principal"].append(
                            " variable: " + identValue)  # Adds "variable" to the dictionary
                    else:
                        # Checks where it has to add the information
                        if (self.functionDecFlag and not self.cycleFlag and not self.conditionalFlag):
                            if (not self.funDecDic[i]["cuerpo funcion"].__contains__("variable: " + identValue)):
                                self.funDecDic[i]["cuerpo funcion"].append(
                                    "variable: " + identValue)

                        if (self.functionDecFlag and self.cycleFlag):
                            if (not self.funDecDic[i]["cuerpo ciclo"].__contains__("variable: " + identValue)):
                                self.funDecDic[i]["cuerpo ciclo"].append(
                                    "variable: " + identValue)

                        if (self.functionDecFlag and self.conditionalFlag):
                            if (not self.funDecDic[i]["cuerpo condicional"].__contains__("variable: " + identValue)):
                                self.funDecDic[i]["cuerpo condicional"].append(
                                    "variable: " + identValue)


def main(argv):
    if len(argv) == 3:
        coutoutfile = argv[2] + ".json"
        output = open(str(coutoutfile), "w")
        fcat = CodeAnalyzer(output)
    else:
        print("Error, instruccion incorrecta.")
        quit()
    inp = FileStream(argv[1])
    lexer = fodcatLexer(inp)
    stream = CommonTokenStream(lexer)
    parser = fodcatParser(stream)
    tree = parser.program()
    walker = ParseTreeWalker()
    walker.walk(fcat, tree)
    output.close()


if __name__ == "__main__":
    main(sys.argv)



####################################################
# Compiler to LIE++ Language
# TEC - FOD
# Authors:
#   Michael Gonzalez Rivera
#   Luis Saborio Hernandez
# Edited:
#   Ellioth Ramirez
####################################################


from analyzer.codeAnalyzer import *
from generator.codeGenerator import *
import sys
import platform
import subprocess

import argparse

sys.path.insert(1, sys.path[0].replace("lang_tools", "gen"))

TRANS_FLAG = "-t"
OUTPUT_EXT = ".py"
OUTPUT_JSON = ".json"
SCRIPT = 1
SOURCE = 0
REVIEW_EXAM = 1

# Verify the device where the compiler is been executed
# @return Boolean value if is running over a Raspberry Pi


def verifyDevice():
    if platform.system() == "Linux":  # If is a Linux OS
        command = "\n cat /proc/cpuinfo"
        all_info = subprocess.check_output(command, shell=True).strip()
        all_info = all_info.decode("utf-8")
        for line in all_info.split("\n"):
            if "model name" in line:
                # If is running over an ARM processor (Raspberry Pi)
                if "ARM" in re.sub(".*model name.*:", "", line, 1):
                    return True
    return False


def main(args):
    canGPIOFlag = verifyDevice()
    mmFun = multiFun()
    if args.de_consola:
        inp = InputStream(mmFun.replaceBadCharacters(args.codigo_fuente))
    else:
        inp = FileStream(args.codigo_fuente)
    lexer = fodcatLexer(inp)
    stream = CommonTokenStream(lexer)
    parser = fodcatParser(stream)

    parser._errHandler = ParserErrorStrategy()
    parser.removeErrorListeners()
    parser.addErrorListener(ParserErrorListener())

    tree = parser.program()
    walker = ParseTreeWalker()

    semanticAnalyzer = SemanticAnalyzer()
    exe_error = False
    try:
        walker.walk(semanticAnalyzer, tree)
    except SystemExit as e:
        exe_error = True
        pass

    if (args.revisar_examen and not exe_error):
        generator = CodeGenerator(None, canGPIOFlag)
        walker.walk(generator, tree)
        print(json.dumps(mmFun.execCode(generator.program)))
        return
    if (args.analisis_estructuras == None and not exe_error):
        outPythonFile = args.resultado_compilado + OUTPUT_EXT
        outPutPythonFile = open(str(outPythonFile), "w")
        generator = CodeGenerator(
            outPutPythonFile, canGPIOFlag, args.modo_entrada, args.entradas)
        walker.walk(generator, tree)
        outPutPythonFile.close()
    elif(args.online):
        generator = CodeGenerator(None, canGPIOFlag)
        if exe_error or len(ParserErrorStrategy.errors) > 0:
            errores = ParserErrorStrategy.errors+SemanticErrorStrategy.errors
            msg_errores = "\n".join(errores)
            response = {"Success": False, "errores": msg_errores}
            ParserErrorStrategy.errors = []
            SemanticErrorStrategy.errors = []
            return response
        walker.walk(generator, tree)
        response = {"Success": True, "codigo": generator.program}
        return response
    elif not exe_error:
        outPythonFile = args.resultado_compilado + OUTPUT_EXT
        outPutPythonFile = open(str(outPythonFile), "w")
        generator = CodeGenerator(
            outPutPythonFile, canGPIOFlag, args.modo_entrada, args.entradas)
        walker.walk(generator, tree)
        outJsonFile = args.analisis_estructuras + OUTPUT_JSON
        outPutJsonFile = open(str(outJsonFile), "w")
        analyzer = CodeAnalyzer(outPutJsonFile)
        walker.walk(analyzer, tree)
        outPutPythonFile.close()
        outPutJsonFile.close()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Compilador para el lenguaje LIE++.')

    parser.add_argument(
        "--codigo-fuente",
        type=str,
        required=True,
        help="Si --de-consola explicitamente usado en la llamada del compilador, este parametro\
            será interpretado como la dirección del archivo con el codigo fuente a compilar,\
            de lo contrario será interpretado como un string con todo el código fuente que se desea compilar"
    )

    parser.add_argument(
        "--de-consola",
        action='store_true',
        required=False,
        help="Indica si el parametro --codigo-fuente debe ser interpretado como una dirección de archivo\
            o como un string con todo el código fuente a compilar."
    )

    parser.add_argument(
        "--revisar-examen",
        action='store_true',
        required=False,
        help="Si esta bandera está presente se realizará una revisión del código como examen."
    )

    parser.add_argument(
        "--resultado-compilado",
        type=str,
        required=True,
        help="Dirección del archivo donde se almacenará el resultado.\
            El compilador usume el archivo de salida es de tipo python, por lo que\
            no se debe agregar la extensión .py explicitamente."
    )

    parser.add_argument(
        "--analisis-estructuras",
        type=str,
        required=False,
        help="Dirección del archivo donde se almacenará el análisis del código.\
            El compilador asume que este archivo es di tipo JSON, po lo que\
            no se debe agregar la extensión .json explicitamente."
    )
    parser.add_argument(
        "--online",
        action="store_true",
        required=False,
        help="Si esta bandera está presente, se compilará de modo online,\
            lo cual provoca que la funcion 'main' retorne un string con el\
            código generado"
    )

    parser.add_argument(
        "--modo-entrada",
        choices=['automatico', 'manual'],
        type=str,
        default="manual",
        required=False,
        help="Cola de entrdas."
    )

    parser.add_argument(
        "--entradas",
        nargs='+',
        required=False,
        help="Cola de entrdas."
    )

    args = parser.parse_args()

    print(args)

    main(args)



####################################################
# LIE++ Parser Error Strategy Handler
# TEC - FOD
# Authors:
#   Michael Gonzalez Rivera
#   Luis Saborio Hernandez
# Edited:
#   Ellioth Ramirez
####################################################

from os import error
import sys

from antlr4 import *
from fodcatParser import fodcatParser
from antlr4.error.ErrorStrategy import DefaultErrorStrategy
from antlr4.error.ErrorListener import ErrorListener
from antlr4.error.Errors import RecognitionException, NoViableAltException, InputMismatchException, \
    FailedPredicateException, ParseCancellationException

# This imports are only needed when testing
if __name__ == '__main__':
    from fodcatListener import fodcatListener
    from fodcatLexer import fodcatLexer


class ParserErrorListener(ErrorListener):
    
    ## Provides a default instance of {@link ConsoleErrorListener}.
    INSTANCE = None

    ##
    # {@inheritDoc}
    #
    # <p>
    # This implementation prints messages to {@link System#err} containing the
    # values of {@code line}, {@code charPositionInLine}, and {@code msg} using
    # the following format.</p>
    #
    # <pre>
    # line <em>line</em>:<em>charPositionInLine</em> <em>msg</em>
    # </pre>
    #
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        print("linea " + str(line) + ":" + str(column) + " " + msg, file=sys.stderr)


class ParserErrorStrategy(DefaultErrorStrategy):
    errors = []

    def __init__(self):
        super().__init__()
        pass

    # This is called by {@link #reportError} when the exception is a
    # {@link NoViableAltException}.
    #
    # @see #reportError
    #
    # @param recognizer the parser instance
    # @param e the recognition exception
    def reportNoViableAlternative(self, recognizer:Parser, e:NoViableAltException):
        tokens = recognizer.getTokenStream()
        if tokens is not None:
            if e.startToken.type==Token.EOF:
                input = "<EOF>"
            else:
                input = tokens.getText(e.startToken, e.offendingToken)
        else:
            input = "<desconocida>"
        msg = "ninguna alternativa valida en la entrada " + \
            self.escapeWSAndQuote(input)
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e.offendingToken, e)

        
    ## This is called by @link ParserErrorStrategy.ParserErrorStrategy.reportFailedPredicate reportFailedPredicate @endlink when the exception is an
    # {@link InputMismatchException}.
    #
    # @see #reportError
    #
    # @param recognizer the parser instance
    # @param e the recognition exception
    #
    def reportInputMismatch(self, recognizer:Parser, e:InputMismatchException):
        msg = "entrada inesperada " + self.getTokenErrorDisplay(e.offendingToken) \
              + " esperando " + e.getExpectedTokens().toString(recognizer.literalNames,
                                                               recognizer.symbolicNames)
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e.offendingToken, e)

        #
    ## This is called by {@link #reportError} when the exception is a
    # {@link FailedPredicateException}.
    #
    # @see #reportError
    # {@ref holis}
    # @param recognizer the parser instance
    # @param e the recognition exception
    #
    def reportFailedPredicate(self, recognizer, e):
        ruleName = recognizer.ruleNames[recognizer._ctx.getRuleIndex()]
        msg = "regla " + ruleName + " " + e.message
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, e.offendingToken, e)

    # This method is called to report a syntax error which requires the removal
    # of a token from the input stream. At the time this method is called, the
    # erroneous symbol is current {@code LT(1)} symbol and has not yet been
    # removed from the input stream. When this method returns,
    # {@code recognizer} is in error recovery mode.
    #
    # <p>This method is called when {@link #singleTokenDeletion} identifies
    # single-token deletion as a viable recovery strategy for a mismatched
    # input error.</p>
    #
    # <p>The default implementation simply returns if the handler is already in
    # error recovery mode. Otherwise, it calls {@link #beginErrorCondition} to
    # enter error recovery mode, followed by calling
    # {@link Parser#notifyErrorListeners}.</p>
    #
    # @param recognizer the parser instance
    #
    def reportUnwantedToken(self, recognizer:Parser):
        if self.inErrorRecoveryMode(recognizer):
            return

        self.beginErrorCondition(recognizer)
        t = recognizer.getCurrentToken()
        tokenName = self.getTokenErrorDisplay(t)
        expecting = self.getExpectedTokens(recognizer)
        msg = "entrada inesperada " + tokenName + " esperando " \
            + expecting.toString(recognizer.literalNames,
                                 recognizer.symbolicNames)
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, t, None)

    # This method is called to report a syntax error which requires the
    # insertion of a missing token into the input stream. At the time this
    # method is called, the missing token has not yet been inserted. When this
    # method returns, {@code recognizer} is in error recovery mode.
    #
    # <p>This method is called when {@link #singleTokenInsertion} identifies
    # single-token insertion as a viable recovery strategy for a mismatched
    # input error.</p>
    #
    # <p>The default implementation simply returns if the handler is already in
    # error recovery mode. Otherwise, it calls {@link #beginErrorCondition} to
    # enter error recovery mode, followed by calling
    # {@link Parser#notifyErrorListeners}.</p>
    #
    # @param recognizer the parser instance
    #
    def reportMissingToken(self, recognizer:Parser):
        if self.inErrorRecoveryMode(recognizer):
            return
        self.beginErrorCondition(recognizer)
        t = recognizer.getCurrentToken()
        expecting = self.getExpectedTokens(recognizer)
        msg = "falta " + expecting.toString(recognizer.literalNames, recognizer.symbolicNames) \
            + " en " + self.getTokenErrorDisplay(t)
        self.errors.append(msg)
        recognizer.notifyErrorListeners(msg, t, None)


# This main method is used to test the module
# 
# @param argv[1] source code file    
def main(argv):
    inp = FileStream(argv[1])
    lexer = fodcatLexer(inp)
    stream = CommonTokenStream(lexer)
    parser = fodcatParser(stream)
    parser._errHandler = ParserErrorStrategy()
    parser.removeErrorListeners()
    parser.addErrorListener(ParserErrorListener())
    tree = parser.program()


if __name__ == '__main__':
    main(sys.argv)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="LIE",
    version="0.1.0",
    author="SoftLab",
    author_email="",
    description="A LIE++ to python compiler",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    # List all your dependencies inside the list
    install_requires=['antlr4-python3-runtime', 'lieGenTools'],
    license='MIT'
)



####################################################
# Compiler to LIE++ Language
# TEC - FOD
# Authors:
#   Ellioth Ramirez
####################################################

import sys, platform, subprocess

sys.path.insert(1, sys.path[0].replace("lang_tools","gen"))

from generator.codeGenerator import *
from analyzer.codeAnalyzer import *
from LIEsemantic.semanticAnalyzer import *
from multi.multiFun import multiFun

TRANS_FLAG = "-t"
OUTPUT_EXT = ".py"
OUTPUT_JSON = ".json"
SCRIPT = 1
SOURCE = 0

## Verify the device where the compiler is been executed
# @return Boolean value if is running over a Raspberry Pi
def verifyDevice():
    if platform.system() == "Linux":  # If is a Linux OS
        command = "\n cat /proc/cpuinfo"
        all_info = subprocess.check_output(command, shell=True).strip()
        all_info = all_info.decode("utf-8")
        for line in all_info.split("\n"):
            if "model name" in line:
                if "ARM" in re.sub(".*model name.*:", "", line,
                                   1):  # If is running over an ARM processor (Raspberry Pi)
                    return True
    return False


def main(argv):
    if (len(argv) < 3):
        print("Error: cantidad de argumentos invalidos")
        quit()
    mmFun = multiFun()
    canGPIOFlag = verifyDevice()
    # This change it's done so we can call the
    # compiler, from the arguments, to a source
    # code (0) or a program file (1).
    if int(argv[1]) == SCRIPT:
        inp = FileStream(argv[2])
    elif int(argv[1]) == SOURCE:
        inp = InputStream(mmFun.replaceBadCharacters(argv[2]))
    lexer = fodcatLexer(inp)
    stream = CommonTokenStream(lexer)
    parser = fodcatParser(stream)

    parser._errHandler = ParserErrorStrategy()
    parser.removeErrorListeners()
    parser.addErrorListener(ParserErrorListener())

    tree = parser.program()
    walker = ParseTreeWalker()

    semanticAnalyzer = SemanticAnalyzer()
    walker.walk(semanticAnalyzer, tree)

    # we're going take chance of the gateway which
    # doesn't write the compiled code to the disk,
    # if the write area it's null/none.
    # If we do this, we still keep the compiled
    # program.
    generator = CodeGenerator(None, canGPIOFlag)
    walker.walk(generator, tree)
    #mmFun.printPreatty(mmFun.execCode(generator.program))
    print(mmFun.execCode(generator.program))

if __name__ == '__main__':
    main(sys.argv)


####################################################
# LIE++ Semantic Error Strategy Handler
# TEC - FOD
# Authors:
#   Michael Gonzalez Rivera
#   Luis Saborio Hernandez
# Edited:
#   Ellioth Ramirez
####################################################

import sys

from LIEsemantic.semanticErrors import *
from LIEsemantic.semanticAnalyzer import SemanticAnalyzer

# Error messeages definition.


class SemanticErrorStrategy():

    # Report an error for an array call to an array wich is no beeing defined.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportArrayNotDefined(self, recognizer: SemanticAnalyzer, e: ArrayNotDefined):
        msg = 'El arreglo: "' + e.offended_identifier + '" no ha sido definido.'
        recognizer.notifyErrorListeners(msg, e)

    # Report an eror to an array call wich do not specify an index.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportArrayWithoutIndex(self, recognizer: SemanticAnalyzer, e: ArrayWithoutIndex):
        msg = 'El arreglo: "' + e.offended_identifier + \
            '" requiere que se indique una posición.'
        recognizer.notifyErrorListeners(msg, e)

    # Report an error for a variable redefinition
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportVariableRedefinition(self, recognizer: SemanticAnalyzer, e: VariableRedefinition):
        msg = 'La variable "' + e.duplicated_identifier + \
            '" fue definida anteriormente. \n'
        recognizer.notifyErrorListeners(msg, e)

    # Report an error for a function redefinition.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportFuctionRedefinition(self, recognizer: SemanticAnalyzer, e: FunctionRedefinition):
        msg = 'Función "' + e.duplicated_identifier + '", con "' + \
            str(e.params_qty) + '" parametros definida anteriormente.'
        recognizer.notifyErrorListeners(msg, e)

    # Report an error for a function call to a function not defined.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportNoMatchedFunction(self, recognizer: SemanticAnalyzer, e: NoMatchedFunction):
        msg = 'Función "' + e.no_matched_identifier + '" con "' + \
            str(e.params_qty) + '" parametros, no se ha definido.'
        recognizer.notifyErrorListeners(msg, e)

    # Report an error when sending no compatible param type.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportNoCompatibleParam(self, recognizer: SemanticAnalyzer, e: NoCompatibleParam):
        msg = 'Paso de parametro incompatible. Tipo esperado "' + e.expected_type + '", tipo recibido "'\
            + e.getted_type + '.'
        recognizer.notifyErrorListeners(msg, e)

    # Report an error for a identifier call wich is not beeing defined.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportNoMatchIdentifier(self, recognizer: SemanticAnalyzer, e: NoMatchIdentifier):
        msg = 'identificador: "' + e.no_matched_identifier + '" no declarado'
        recognizer.notifyErrorListeners(msg, e)

    # Depending of the semantic error type chose the error handler.
    # @param recognizer Object wich have the error listener list.
    # @param e Data error object.
    def reportSemanticError(self, recognizer: SemanticAnalyzer, e: SemanticError):
        if isinstance(e, NoMatchIdentifier):
            self.reportNoMatchIdentifier(recognizer, e)
        elif isinstance(e, FunctionRedefinition):
            self.reportFuctionRedefinition(recognizer, e)
        elif isinstance(e, NoMatchedFunction):
            self.reportNoMatchedFunction(recognizer, e)
        elif isinstance(e, VariableRedefinition):
            self.reportVariableRedefinition(recognizer, e)
        elif isinstance(e, ArrayWithoutIndex):
            self.reportArrayWithoutIndex(recognizer, e)
        elif isinstance(e, ArrayNotDefined):
            self.reportArrayNotDefined(recognizer, e)
        elif isinstance(e, NoCompatibleParam):
            self.reportNoCompatibleParam(recognizer, e)

        sys.exit(0)
